import numpy as np
import pandas as pd
import scipy as sp

from typing import Optional, Sequence, Union
from anndata import AnnData

MIN_VERSION = "0.6.5"


def scvi(
    adata: AnnData,
    n_hidden: int = 128,
    n_latent: int = 10,
    n_layers: int = 1,
    dispersion: str = "gene",
    n_epochs: int = 400,
    lr: int = 1e-3,
    train_size: int = 1.0,
    batch_key: Optional[str] = None,
    use_highly_variable_features: bool = True,
    subset_features: Optional[Sequence[Union[int, str]]] = None,
    linear_decoder: bool = False,
    copy: bool = False,
    use_cuda: bool = True,
    return_posterior: bool = True,
    trainer_kwargs: dict = {},
    model_kwargs: dict = {},
) -> Optional[AnnData]:
    """\
    SCVI [Lopez18]_.
    Fits scVI model onto raw count data given an anndata object
    scVI uses stochastic optimization and deep neural networks to aggregate information 
    across similar subjects and features and to approximate the distributions that underlie
    observed expression values, while accounting for batch effects and limited sensitivity.
    To use a linear-decoded Variational AutoEncoder model (implementation of [Svensson20]_.),
    set linear_decoded = True. Compared to standard VAE, this model is less powerful, but can 
    be used to inspect which features contribute to variation in the dataset. It may also be used
    for all scVI tasks, like differential expression, batch correction, imputation, etc.
    However, batch correction may be less powerful as it assumes a linear model.
    .. note::
        More information and bug reports `here <https://github.com/YosefLab/scVI>`__.
    Parameters
    ----------
    adata
        An anndata file with `X` attribute of unnormalized count data
    n_hidden
        Number of nodes per hidden layer
    n_latent
        Dimensionality of the latent space
    n_layers
        Number of hidden layers used for encoder and decoder NNs
    dispersion
        One of the following
        * `'gene'` - dispersion parameter of NB is constant per gene(feature) across cells(subjects)
        * `'gene-batch'` - dispersion can differ between different batches
        * `'gene-label'` - dispersion can differ between different labels
        * `'gene-cell'` - dispersion can differ for every gene(feature) in every cell(subject)
    n_epochs
        Number of epochs to train
    lr
        Learning rate
    train_size
        The train size, either a float between 0 and 1 or an integer for the number of training samples to use
    batch_key
        Column name in anndata.obs for batches. 
        If None, no batch correction is performed
        If not None, batch correction is performed per batch category
    use_highly_variable_features
        If true, uses only the features in anndata.var["highly_variable"]
    subset_features
        Optional list of indices or feature names to subset anndata. 
        If not None, use_highly_variable_features is ignored
    linear_decoder
        If true, uses LDVAE model, which is an implementation of [Svensson20]_.
    copy
        If true, a copy of anndata is returned
    return_posterior
        If true, posterior object is returned
    use_cuda
        If true, uses cuda
    trainer_kwargs
        Extra arguments for UnsupervisedTrainer
    model_kwargs
        Extra arguments for VAE or LDVAE model
    
    Returns
    -------
    If `copy` is true, anndata is returned.
    If `return_posterior` is true, the posterior object is returned
    If both `copy` and `return_posterior` are true, 
    a tuple of anndata and the posterior are returned in that order. 
    `adata.obsm['X_scvi']` stores the latent representations
    `adata.obsm['X_scvi_denoised']` stores the normalized mean of the negative binomial
    `adata.obsm['X_scvi_sample_rate']` stores the mean of the negative binomial
    
    If linear_decoder is true:
    `adata.uns['ldvae_loadings']` stores the per-gene weights in the linear decoder as a
    genes by n_latent matrix.
    """

    try:
        from scvi.models import VAE, LDVAE
        from scvi.inference import UnsupervisedTrainer
        from scvi.dataset import AnnDatasetFromAnnData
    except ImportError:
        raise ImportError(
            "Please install scvi package from https://github.com/YosefLab/scVI"
        )

    # check if observations are unnormalized using first 10
    # code from: https://github.com/theislab/dca/blob/89eee4ed01dd969b3d46e0c815382806fbfc2526/dca/io.py#L63-L69
    if len(adata) > 10:
        X_subset = adata.X[:10]
    else:
        X_subset = adata.X
    norm_error = (
        'Make sure that the dataset (adata.X) contains unnormalized count data.'
    )
    if sp.sparse.issparse(X_subset):
        assert (X_subset.astype(int) != X_subset).nnz == 0, norm_error
    else:
        assert np.all(X_subset.astype(int) == X_subset), norm_error

    if subset_features is not None:
        adata_subset = adata[:, subset_features]
    elif use_highly_variable_features and "highly_variable" in adata.var:
        adata_subset = adata[:, adata.var["highly_variable"]]
    else:
        adata_subset = adata

    if batch_key is not None:
        codes, uniques = pd.factorize(adata_subset.obs[batch_key])
        adata_subset.obs['_tmp_scvi_batch'] = codes
        n_batches = len(uniques)
    else:
        n_batches = 0

    dataset = AnnDatasetFromAnnData(adata_subset.copy(), batch_label='_tmp_scvi_batch')

    if linear_decoder:
        vae = LDVAE(
            n_input=dataset.nb_genes,
            n_batch=n_batches,
            n_labels=dataset.n_labels,
            n_hidden=n_hidden,
            n_latent=n_latent,
            n_layers_encoder=n_layers,
            dispersion=dispersion,
            **model_kwargs,
        )

    else:
        vae = VAE(
            dataset.nb_genes,
            n_batch=n_batches,
            n_labels=dataset.n_labels,
            n_hidden=n_hidden,
            n_latent=n_latent,
            n_layers=n_layers,
            dispersion=dispersion,
            **model_kwargs,
        )

    trainer = UnsupervisedTrainer(
        model=vae,
        gene_dataset=dataset,
        use_cuda=use_cuda,
        train_size=train_size,
        **trainer_kwargs,
    )

    trainer.train(n_epochs=n_epochs, lr=lr)

    full = trainer.create_posterior(
        trainer.model, dataset, indices=np.arange(len(dataset))
    )
    latent, batch_indices, labels = full.sequential().get_latent()

    if copy:
        adata = adata.copy()

    adata.obsm['X_scvi'] = latent
    adata.obsm['X_scvi_denoised'] = full.sequential().get_sample_scale()
    adata.obsm['X_scvi_sample_rate'] = full.sequential().imputation()

    if linear_decoder:
        loadings = vae.get_loadings()
        df = pd.DataFrame(loadings, index=adata_subset.var_names)
        adata.uns['ldvae_loadings'] = df

    if copy and return_posterior:
        return adata, full
    elif copy:
        return adata
    elif return_posterior:
        return full



from typing import Optional, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as pl
from matplotlib import rcParams
from anndata import AnnData
from . import _utils

# --------------------------------------------------------------------------------
# Plot result of preprocessing functions
# --------------------------------------------------------------------------------


def highly_variable_features(
    adata_or_result: Union[AnnData, pd.DataFrame, np.recarray],
    log: bool = False,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
    highly_variable_features: bool = True,
):
    """Plot dispersions or normalized variance versus means for features.

    Produces Supp. Fig. 5c of Zheng et al. (2017) and MeanVarPlot() and
    VariableFeaturePlot() of Seurat.

    Parameters
    ----------
    adata
        Result of :func:`~quanp.pp.highly_variable_features`.
    log
        Plot on logarithmic axes.
    show
         Show the plot, do not return axis.
    save
        If `True` or a `str`, save the figure.
        A string is appended to the default filename.
        Infer the filetype if ending on {{`'.pdf'`, `'.png'`, `'.svg'`}}.
    """
    if isinstance(adata_or_result, AnnData):
        result = adata_or_result.var
        seurat_v3_flavor = adata_or_result.uns["hvg"]["flavor"] == "seurat_v3"
    else:
        result = adata_or_result
        if isinstance(result, pd.DataFrame):
            seurat_v3_flavor = "variances_norm" in result.columns
        else:
            seurat_v3_flavor = False
    if highly_variable_features:
        feature_subset = result.highly_variable
    else:
        feature_subset = result.feature_subset
    means = result.means

    if seurat_v3_flavor:
        var_or_disp = result.variances
        var_or_disp_norm = result.variances_norm
    else:
        var_or_disp = result.dispersions
        var_or_disp_norm = result.dispersions_norm
    size = rcParams['figure.figsize']
    pl.figure(figsize=(2 * size[0], size[1]))
    pl.subplots_adjust(wspace=0.3)
    for idx, d in enumerate([var_or_disp_norm, var_or_disp]):
        pl.subplot(1, 2, idx + 1)
        for label, color, mask in zip(
            ['highly variable features', 'other features'],
            ['black', 'grey'],
            [feature_subset, ~feature_subset],
        ):
            if False:
                means_, var_or_disps_ = np.log10(means[mask]), np.log10(d[mask])
            else:
                means_, var_or_disps_ = means[mask], d[mask]
            pl.scatter(means_, var_or_disps_, label=label, c=color, s=1)
        if log:  # there's a bug in autoscale
            pl.xscale('log')
            pl.yscale('log')
            y_min = np.min(var_or_disp)
            y_min = 0.95 * y_min if y_min > 0 else 1e-1
            pl.xlim(0.95 * np.min(means), 1.05 * np.max(means))
            pl.ylim(y_min, 1.05 * np.max(var_or_disp))
        if idx == 0:
            pl.legend()
        pl.xlabel(('$log_{10}$ ' if False else '') + 'mean expressions of features')
        data_type = 'dispersions' if not seurat_v3_flavor else 'variances'
        pl.ylabel(
            ('$log_{10}$ ' if False else '')
            + '{} of features'.format(data_type)
            + (' (normalized)' if idx == 0 else ' (not normalized)')
        )

    _utils.savefig_or_show('filter_features_dispersion', show=show, save=save)
    if show is False:
        return pl.gca()


# backwards compat
def filter_features_dispersion(
    result: np.recarray,
    log: bool = False,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
):
    """\
    Plot dispersions versus means for features.

    Produces Supp. Fig. 5c of Zheng et al. (2017) and MeanVarPlot() of Seurat.

    Parameters
    ----------
    result
        Result of :func:`~quanp.pp.filter_features_dispersion`.
    log
        Plot on logarithmic axes.
    show
         Show the plot, do not return axis.
    save
        If `True` or a `str`, save the figure.
        A string is appended to the default filename.
        Infer the filetype if ending on {{`'.pdf'`, `'.png'`, `'.svg'`}}.
    """
    highly_variable_features(
        result, log=log, show=show, save=save, highly_variable_features=False
    )



from types import MappingProxyType
from typing import Optional, Tuple, Sequence, Type, Mapping, Any

import numpy as np
import pandas as pd
from anndata import AnnData
from natsort import natsorted
from scipy.sparse import spmatrix
from packaging import version

from ._utils_clustering import rename_groups, restrict_adjacency
from .. import _utils, logging as logg
from .._compat import Literal
from .._utils import _choose_graph

try:
    from louvain.VertexPartition import MutableVertexPartition
except ImportError:
    class MutableVertexPartition: pass
    MutableVertexPartition.__module__ = 'louvain.VertexPartition'


def louvain(
    adata: AnnData,
    resolution: Optional[float] = None,
    random_state: _utils.AnyRandom = 0,
    restrict_to: Optional[Tuple[str, Sequence[str]]] = None,
    key_added: str = 'louvain',
    adjacency: Optional[spmatrix] = None,
    flavor: Literal['vtraag', 'igraph', 'rapids'] = 'vtraag',
    directed: bool = True,
    use_weights: bool = False,
    partition_type: Optional[Type[MutableVertexPartition]] = None,
    partition_kwargs: Mapping[str, Any] = MappingProxyType({}),
    neighbors_key: Optional[str] = None,
    obsp: Optional[str] = None,
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Cluster subjects into subgroups [Blondel08]_ [Levine15]_ [Traag17]_.

    Cluster subjects using the Louvain algorithm [Blondel08]_ in the implementation
    of [Traag17]_. The Louvain algorithm has been proposed for subject
    analysis by [Levine15]_.

    This requires having ran :func:`~quanp.pp.neighbors` or
    :func:`~quanp.external.pp.bbknn` first,
    or explicitly passing a ``adjacency`` matrix.

    Parameters
    ----------
    adata
        The annotated data matrix.
    resolution
        For the default flavor (``'vtraag'``), you can provide a resolution
        (higher resolution means finding more and smaller clusters),
        which defaults to 1.0.
        See “Time as a resolution parameter” in [Lambiotte09]_.
    random_state
        Change the initialization of the optimization.
    restrict_to
        Restrict the clustering to the categories within the key for sample
        annotation, tuple needs to contain ``(obs_key, list_of_categories)``.
    key_added
        Key under which to add the cluster labels. (default: ``'louvain'``)
    adjacency
        Sparse adjacency matrix of the graph, defaults to neighbors connectivities.
    flavor
        Choose between to packages for computing the clustering.
        ``'vtraag'`` is much more powerful, and the default.
    directed
        Interpret the ``adjacency`` matrix as directed graph?
    use_weights
        Use weights from knn graph.
    partition_type
        Type of partition to use.
        Only a valid argument if ``flavor`` is ``'vtraag'``.
    partition_kwargs
        Key word arguments to pass to partitioning,
        if ``vtraag`` method is being used.
    neighbors_key
        Use neighbors connectivities as adjacency.
        If not specified, louvain looks .obsp['connectivities'] for connectivities
        (default storage place for pp.neighbors).
        If specified, louvain looks
        .obsp[.uns[neighbors_key]['connectivities_key']] for connectivities.
    obsp
        Use .obsp[obsp] as adjacency. You can't specify both
        `obsp` and `neighbors_key` at the same time.
    copy
        Copy adata or modify it inplace.

    Returns
    -------
    :obj:`None`
        By default (``copy=False``), updates ``adata`` with the following fields:

        ``adata.obs['louvain']`` (:class:`pandas.Series`, dtype ``category``)
            Array of dim (number of samples) that stores the subgroup id
            (``'0'``, ``'1'``, ...) for each subject.

    :class:`~anndata.AnnData`
        When ``copy=True`` is set, a copy of ``adata`` with those fields is returned.
    """
    partition_kwargs = dict(partition_kwargs)
    start = logg.info('running Louvain clustering')
    if (flavor != 'vtraag') and (partition_type is not None):
        raise ValueError(
            '`partition_type` is only a valid argument '
            'when `flavour` is "vtraag"'
        )
    adata = adata.copy() if copy else adata
    if adjacency is None:
        adjacency = _choose_graph(adata, obsp, neighbors_key)
    if restrict_to is not None:
        restrict_key, restrict_categories = restrict_to
        adjacency, restrict_indices = restrict_adjacency(
            adata,
            restrict_key,
            restrict_categories,
            adjacency,
        )
    if flavor in {'vtraag', 'igraph'}:
        if flavor == 'igraph' and resolution is not None:
            logg.warning(
                '`resolution` parameter has no effect for flavor "igraph"'
            )
        if directed and flavor == 'igraph':
            directed = False
        if not directed: logg.debug('    using the undirected graph')
        g = _utils.get_igraph_from_adjacency(adjacency, directed=directed)
        if use_weights:
            weights = np.array(g.es["weight"]).astype(np.float64)
        else:
            weights = None
        if flavor == 'vtraag':
            import louvain
            if partition_type is None:
                partition_type = louvain.RBConfigurationVertexPartition
            if resolution is not None:
                partition_kwargs["resolution_parameter"] = resolution
            if use_weights:
                partition_kwargs["weights"] = weights
            if version.parse(louvain.__version__) < version.parse("0.7.0"):
                louvain.set_rng_seed(random_state)
            else:
                partition_kwargs["seed"] = random_state
            logg.info('    using the "louvain" package of Traag (2017)')
            part = louvain.find_partition(
                g, partition_type,
                **partition_kwargs,
            )
            # adata.uns['louvain_quality'] = part.quality()
        else:
            part = g.community_multilevel(weights=weights)
        groups = np.array(part.membership)
    elif flavor == 'rapids':
        # nvLouvain only works with undirected graphs,
        # and `adjacency` must have a directed edge in both directions
        import cudf
        import cugraph
        offsets = cudf.Series(adjacency.indptr)
        indices = cudf.Series(adjacency.indices)
        if use_weights:
            sources, targets = adjacency.nonzero()
            weights = adjacency[sources, targets]
            if isinstance(weights, np.matrix):
                weights = weights.A1
            weights = cudf.Series(weights)
        else:
            weights = None
        g = cugraph.Graph()
        g.add_adj_list(offsets, indices, weights)
        logg.info('    using the "louvain" package of rapids')
        louvain_parts, _ = cugraph.louvain(g)
        groups = louvain_parts.to_pandas().sort_values('vertex')[['partition']].to_numpy().ravel()
    elif flavor == 'taynaud':
        # this is deprecated
        import networkx as nx
        import community
        g = nx.Graph(adjacency)
        partition = community.best_partition(g)
        groups = np.zeros(len(partition), dtype=int)
        for k, v in partition.items(): groups[k] = v
    else:
        raise ValueError(
            '`flavor` needs to be "vtraag" or "igraph" or "taynaud".'
        )
    if restrict_to is not None:
        if key_added == 'louvain':
            key_added += '_R'
        groups = rename_groups(
            adata,
            key_added,
            restrict_key,
            restrict_categories,
            restrict_indices,
            groups,
        )
    adata.obs[key_added] = pd.Categorical(
        values=groups.astype('U'),
        categories=natsorted(map(str, np.unique(groups))),
    )
    adata.uns['louvain'] = {}
    adata.uns['louvain']['params'] = dict(
        resolution=resolution,
        random_state=random_state,
    )
    logg.info(
        '    finished',
        time=start,
        deep=(
            f'found {len(np.unique(groups))} clusters and added\n'
            f'    {key_added!r}, the cluster labels (adata.obs, categorical)'
        ),
    )
    return adata if copy else None



from typing import Optional, Union, Iterable, Dict

import numpy as np
from anndata import AnnData
from scipy.sparse import issparse
from sklearn.utils import sparsefuncs

from .. import logging as logg
from .._compat import Literal
from .._utils import view_to_actual


def _normalize_data(X, counts, after=None, copy=False):
    X = X.copy() if copy else X
    if issubclass(X.dtype.type, (int, np.integer)):
        X = X.astype(np.float32)  # TODO: Check if float64 should be used
    counts = np.asarray(counts)  # dask doesn't do medians
    after = np.median(counts[counts>0], axis=0) if after is None else after
    counts += (counts == 0)
    counts = counts / after
    if issparse(X):
        sparsefuncs.inplace_row_scale(X, 1/counts)
    else:
        np.divide(X, counts[:, None], out=X)
    return X


def normalize_total(
    adata: AnnData,
    target_sum: Optional[float] = None,
    exclude_highly_expressed: bool = False,
    max_fraction: float = 0.05,
    key_added: Optional[str] = None,
    layers: Union[Literal['all'], Iterable[str]] = None,
    layer_norm: Optional[str] = None,
    inplace: bool = True,
) -> Optional[Dict[str, np.ndarray]]:
    """\
    Normalize counts per subject.
    If choosing `target_sum=1e6`, this is CPM normalization.
    If `exclude_highly_expressed=True`, very highly expressed features are excluded
    from the computation of the normalization factor (size factor) for each
    subject. This is meaningful as these can strongly influence the resulting
    normalized values for all other features [Weinreb17]_.
    Similar functions are used, for example, by Seurat [Satija15]_, Cell Ranger
    [Zheng17]_ or SPRING [Weinreb17]_.
    Params
    ------
    adata
        The annotated data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to subjects and columns to features.
    target_sum
        If `None`, after normalization, each observation (subject) has a total
        count equal to the median of total counts for observations (subjects)
        before normalization.
    exclude_highly_expressed
        Exclude (very) highly expressed features for the computation of the
        normalization factor (size factor) for each subject. A feature is considered
        highly expressed, if it has more than `max_fraction` of the total counts
        in at least one subject. The not-excluded features will sum up to
        `target_sum`.
    max_fraction
        If `exclude_highly_expressed=True`, consider subjects as highly expressed
        that have more counts than `max_fraction` of the original total counts
        in at least one subject.
    key_added
        Name of the field in `adata.obs` where the normalization factor is
        stored.
    layers
        List of layers to normalize. Set to `'all'` to normalize all layers.
    layer_norm
        Specifies how to normalize layers:
        * If `None`, after normalization, for each layer in *layers* each subject
          has a total count equal to the median of the *counts_per_subject* before
          normalization of the layer.
        * If `'after'`, for each layer in *layers* each subject has
          a total count equal to `target_sum`.
        * If `'X'`, for each layer in *layers* each subject has a total count
          equal to the median of total counts for observations (subjects) of
          `adata.X` before normalization.
    inplace
        Whether to update `adata` or return dictionary with normalized copies of
        `adata.X` and `adata.layers`.
    Returns
    -------
    Returns dictionary with normalized copies of `adata.X` and `adata.layers`
    or updates `adata` with normalized version of the original
    `adata.X` and `adata.layers`, depending on `inplace`.
    Example
    --------
    >>> from anndata import AnnData
    >>> import scanpy as sc
    >>> sc.settings.verbosity = 2
    >>> np.set_printoptions(precision=2)
    >>> adata = AnnData(np.array([
    ...    [3, 3, 3, 6, 6],
    ...    [1, 1, 1, 2, 2],
    ...    [1, 22, 1, 2, 2],
    ... ]))
    >>> adata.X
    array([[ 3.,  3.,  3.,  6.,  6.],
           [ 1.,  1.,  1.,  2.,  2.],
           [ 1., 22.,  1.,  2.,  2.]], dtype=float32)
    >>> X_norm = sc.pp.normalize_total(adata, target_sum=1, inplace=False)['X']
    >>> X_norm
    array([[0.14, 0.14, 0.14, 0.29, 0.29],
           [0.14, 0.14, 0.14, 0.29, 0.29],
           [0.04, 0.79, 0.04, 0.07, 0.07]], dtype=float32)
    >>> X_norm = sc.pp.normalize_total(
    ...     adata, target_sum=1, exclude_highly_expressed=True,
    ...     max_fraction=0.2, inplace=False
    ... )['X']
    The following highly-expressed features are not considered during normalization factor computation:
    ['1', '3', '4']
    >>> X_norm
    array([[ 0.5,  0.5,  0.5,  1. ,  1. ],
           [ 0.5,  0.5,  0.5,  1. ,  1. ],
           [ 0.5, 11. ,  0.5,  1. ,  1. ]], dtype=float32)
    """
    if max_fraction < 0 or max_fraction > 1:
        raise ValueError('Choose max_fraction between 0 and 1.')

    if layers == 'all':
        layers = adata.layers.keys()
    elif isinstance(layers, str):
        raise ValueError(
            f"`layers` needs to be a list of strings or 'all', not {layers!r}"
        )

    view_to_actual(adata)

    feature_subset = None
    msg = 'normalizing counts per subject'
    if exclude_highly_expressed:
        counts_per_subject = adata.X.sum(1)  # original counts per subject
        counts_per_subject = np.ravel(counts_per_subject)

        # at least one subject as more than max_fraction of counts per subject
        feature_subset = (adata.X > counts_per_subject[:, None]*max_fraction).sum(0)
        feature_subset = (np.ravel(feature_subset) == 0)

        msg += (
            ' The following highly-expressed features are not considered during '
            f'normalization factor computation:\n{adata.var_names[~feature_subset].tolist()}'
        )
    start = logg.info(msg)

    # counts per subject for subset, if max_fraction!=1
    X = adata.X if feature_subset is None else adata[:, feature_subset].X
    counts_per_subject = X.sum(1)
    # get rid of adata view
    counts_per_subject = np.ravel(counts_per_subject).copy()

    subject_subset = counts_per_subject > 0
    if not np.all(subject_subset):
        logg.warning('Some subjects have total count of features equal to zero')

    if layer_norm == 'after':
        after = target_sum
    elif layer_norm == 'X':
        after = np.median(counts_per_subject[subject_subset])
    elif layer_norm is None:
        after = None
    else:
        raise ValueError('layer_norm should be "after", "X" or None')
    del subject_subset

    if inplace:
        if key_added is not None:
            adata.obs[key_added] = counts_per_subject
        adata.X = _normalize_data(adata.X, counts_per_subject, target_sum)
    else:
        # not recarray because need to support sparse
        dat = dict(
            X=_normalize_data(adata.X, counts_per_subject, target_sum, copy=True),
            norm_factor=counts_per_subject,
        )

    for layer_name in (layers or ()):
        layer = adata.layers[layer_name]
        counts = np.ravel(layer.sum(1))
        if inplace:
            adata.layers[layer_name] = _normalize_data(layer, counts, after)
        else:
            dat[layer_name] = _normalize_data(layer, counts, after, copy=True)

    logg.info(
        '    finished ({time_passed})',
        time=start,
    )
    if key_added is not None:
        logg.debug(f'and added {key_added!r}, counts per subject before normalization (adata.obs)')

    return dat if not inplace else None



"""Set the default matplotlib.rcParams.
"""

import matplotlib
from matplotlib import rcParams
from cycler import cycler

from . import palettes


def set_rcParams_quanp(fontsize=14, color_map=None):
    """Set matplotlib.rcParams to Quanpy defaults.

    Call this through `settings.set_figure_params`.
    """

    # figure
    rcParams['figure.figsize'] = (4, 4)
    rcParams['figure.subplot.left'] = 0.18
    rcParams['figure.subplot.right'] = 0.96
    rcParams['figure.subplot.bottom'] = 0.15
    rcParams['figure.subplot.top'] = 0.91

    rcParams['lines.linewidth'] = 1.5  # the line width of the frame
    rcParams['lines.markersize'] = 6
    rcParams['lines.markeredgewidth'] = 1

    # font
    rcParams['font.sans-serif'] = [
        'Arial',
        'Helvetica',
        'DejaVu Sans',
        'Bitstream Vera Sans',
        'sans-serif',
    ]
    fontsize = fontsize
    rcParams['font.size'] = fontsize
    rcParams['legend.fontsize'] = 0.92 * fontsize
    rcParams['axes.titlesize'] = fontsize
    rcParams['axes.labelsize'] = fontsize

    # legend
    rcParams['legend.numpoints'] = 1
    rcParams['legend.scatterpoints'] = 1
    rcParams['legend.handlelength'] = 0.5
    rcParams['legend.handletextpad'] = 0.4

    # color cycle
    rcParams['axes.prop_cycle'] = cycler(color=palettes.default_20)

    # lines
    rcParams['axes.linewidth'] = 0.8
    rcParams['axes.edgecolor'] = 'black'
    rcParams['axes.facecolor'] = 'white'

    # ticks
    rcParams['xtick.color'] = 'k'
    rcParams['ytick.color'] = 'k'
    rcParams['xtick.labelsize'] = fontsize
    rcParams['ytick.labelsize'] = fontsize

    # axes grid
    rcParams['axes.grid'] = True
    rcParams['grid.color'] = '.8'

    # color map
    rcParams['image.cmap'] = rcParams['image.cmap'] if color_map is None else color_map


def set_rcParams_defaults():
    """Reset `matplotlib.rcParams` to defaults."""
    rcParams.update(matplotlib.rcParamsDefault)



from pathlib import Path
from typing import Optional
import logging, lxml, os, requests, sys, time, urllib, warnings

from bs4 import BeautifulSoup
from tqdm import tqdm

import numpy as np
import pandas as pd
import pandas_market_calendars as mcal
import itertools
import glob

from pandas.io.json import json_normalize
from anndata import AnnData

from .. import logging as logg, _utils
from .._compat import Literal
from .._settings import settings
from ..readwrite import read
from ._utils import check_datasetdir_exists, check_logdir_exists

# # import program settings
# sys.path.append('../../settings')  
from ..local_settings import TOS_API_KEY

HERE = Path(__file__).parent


def blobs(
    n_variables: int = 11,
    n_centers: int = 5,
    cluster_std: float = 1.0,
    n_observations: int = 640,
) -> AnnData:
    """\
    Gaussian Blobs.
    Parameters
    ----------
    n_variables
        Dimension of feature space.
    n_centers
        Number of cluster centers.
    cluster_std
        Standard deviation of clusters.
    n_observations
        Number of observations. By default, this is the same observation number
        as in :func:`scanpy.datasets.krumsiek11`.
    Returns
    -------
    Annotated data matrix containing a observation annotation 'blobs' that
    indicates cluster identity.
    """
    import sklearn.datasets

    X, y = sklearn.datasets.make_blobs(
        n_samples=n_observations,
        n_features=n_variables,
        centers=n_centers,
        cluster_std=cluster_std,
        random_state=0,
    )
    return AnnData(X, obs=dict(blobs=y.astype(str)))


########################################################################################

@check_datasetdir_exists
def wiki_sp500_members_update(update_filename) -> pd.DataFrame:
    """\
    A function to update Standard & Poor 500 listed companies according to wikipedia
    @ https://en.wikipedia.org/wiki/List_of_S%26P_500_companies
    """
    template_filename = HERE / 'sp500.csv'
    update_filename = HERE / update_filename
    verbosity_save = settings.verbosity
    settings.verbosity = 'error'  # suppress output...
    settings.verbosity = verbosity_save

    df = pd.read_csv(template_filename)
    df['date'] = pd.to_datetime(df['date'], dayfirst=True )
    df.set_index('date', inplace=True, drop=True)

    df_update = pd.read_csv(update_filename)
    df_update['date'] = pd.to_datetime(df_update['date'])
    df_update = df_update.sort_values(by=['date'], ascending=True)
    df_update.set_index('date', inplace=True, drop=True)

    df_dict = {}
    for item in df.index:
        df_dict[item.strftime('%Y-%m-%d')] = df.at[item, 'tickers'].split(',')
    
    last_date = df.index[-1:].strftime('%Y-%m-%d')[0]
    df_update.index = df_update.index.strftime('%Y-%m-%d')

    for dt in df_update.index.unique():
        df_dict[dt] = df_dict[last_date].copy()
    
        try:
            if df_update.at[dt, 'ticker_add']:
                tick = df_update.at[dt, 'ticker_add']
                df_dict[dt].append(tick)
                print('Added', df_update.at[dt, 'ticker_add'])
            
        except:
            for tick in df_update.at[dt, 'ticker_add']:
                if tick is not np.nan:
                    df_dict[dt].append(tick)
                    print('Added', tick)
                
        try:
            if df_update.at[dt, 'ticker_drop']:
                df_dict[dt].remove(df_update.at[dt, 'ticker_drop'])
                print('Dropped', df_update.at[dt, 'ticker_drop'])

        except:
            for tick in df_update.at[dt, 'ticker_drop']:
                if tick is not np.nan:
                    df_dict[dt].remove(tick)
                    print('Dropped', tick)
                
        last_date = dt

    df_final = pd.DataFrame(columns=['tickers'])

    for key in df_dict.keys():
        df_final.at[key, 'tickers'] = ','.join(df_dict[key])

    df_final.to_csv(HERE / 'sp500_updated.csv')

    return df_final



@check_datasetdir_exists
@check_logdir_exists
def download_tickers_price_history():
    """
    A function to download price history of Standard & Poor 500 listed companies based 
    on the updated sp500 members. SPY will also be downloaded as the
    benchmark.
    """
    # setting logger 
    logging.basicConfig(filename= settings.logdir / 'price_history_download.txt', 
                        level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(message)s')

    # 1. the input csv directory
    csv_dir = HERE / 'sp500_updated.csv'

    # 2. create output directory if not exist
    output_dir = settings.datasetdir / 'daily'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # the daily eod price
    try:
        df_meta = pd.read_csv(csv_dir, index_col=0, parse_dates=[0])
        df_meta_selected = df_meta.loc[df_meta.index > pd.Timestamp('2000-8-05')]
        tickers_list = list(df_meta_selected['tickers'].values)
        tickers_list_2d = [ls.split(',') for ls in tickers_list]
        tickers = list(itertools.chain.from_iterable(tickers_list_2d))
        tickers = set(tickers + ['SPY'])
        print('Total tickers to download:', len(tickers))

    except KeyError as ke:
        print(KeyError, 'Please provide a csv that contain a column', 
                        'named {}'.format(ke))
        logging.info('Please provide a csv that contain a column \
                        named {}'.format(ke))
        exit(1)


    # define current datetime in miliseconds
    current_milli_time = lambda: int(round(time.time() * 1000))

    # define payload
    payload = {'apikey': TOS_API_KEY,
                'periodType': 'year',
                'frequencyType': 'daily',
                'frequency': '1',
                'period': '20',
                'endDate': current_milli_time()}

    # Match with NYSE calendar
    nyse = mcal.get_calendar('NYSE')
    schedule = nyse.schedule(start_date='1996-01-01', end_date='2022-12-31')

    tickers_failed = []

    # loop and process through each provided ticker (including SPY)
    for ticker in tqdm(tickers, desc='Downloading and processing price history...', 
                                unit='ticker', 
                                position=0):

        # define the endpoint
        endpoint = "https://api.tdameritrade.com/v1/marketdata/{}/pricehistory".format(ticker)

        # request and process data
        content = requests.get(url = endpoint, params = payload)
        data = content.json()

        try:

            # dataframing
            df_tmp = pd.DataFrame.from_dict(data['candles'])
            df_tmp['datetime'] = pd.to_datetime(df_tmp['datetime'], unit='ms').dt.date
            df_tmp = df_tmp[['datetime', 'open', 'high', 'low', 'close', 'volume']]
            df_tmp.set_index('datetime', inplace=True, drop=True)
            df_tmp = df_tmp[df_tmp.index.astype(str).isin(schedule.index.astype(str))]
        
            # Reindex the df to accommodate missing values at certain NYSE calendar
            schedule_order = schedule.loc[df_tmp.index[0]:df_tmp.index[-1]]
            df_tmp = df_tmp.reindex(schedule_order.index) 
        
            # save the outputs
            if not len(df_tmp) < 2:
                df_tmp.to_csv(output_dir / '{}.csv'.format(ticker), sep=',', mode='w')
                print('Downloaded data up to {} for {}'.format(df_tmp.index[-1:][0], ticker))
            
            else:
                print('Total price history collected for {} is < 2 - download \
                     discarded'.format(ticker))
                tickers_failed.append(ticker)                

        except KeyError as ke:
            print(KeyError, 'column {} not found in the processed dataframe.'.format(ke), 
                'Possible reason may be the ticker provided not found in TOS database.')
            tickers_failed.append(ticker)

    logging.info('column {} not found in the processed dataframe. Possible \
                reason may be the ticker provided not found in TOS \
                database.'.format(tickers_failed))


    # list all downloaded tickers 
    downloaded_tickers = os.listdir(output_dir)

    # collect list of csv with at least one na or negative value
    ls_csv_with_na_or_neg = []
    for csv in downloaded_tickers:

        df = pd.read_csv(output_dir / csv, index_col=0)

        for col in df.columns:
            df[col] = pd.to_numeric(df[col], errors='coerce')
        
        if df.isnull().values.any() or df.lt(0).any().any():
            ls_csv_with_na_or_neg.append(csv)

    # remove csv with na or negative value
    for csv in ls_csv_with_na_or_neg:
        os.remove(output_dir / csv)

    # # filter out downloaded tickers (only for use when data downloaded)
    # downloaded_tickers = set([os.path.splitext(f)[0] for f in os.listdir(output_dir)])
    # tickers_failed = tickers - downloaded_tickers

    # matching the sp500 members for analyses
    df_dict = {}
    for item in df_meta_selected.index:
        df_dict[item.strftime('%Y-%m-%d')] = df_meta_selected.at[item, 'tickers'].split(',')
    
    for ticker in tickers_failed:
        for key in df_dict.keys():
            try:
                df_dict[key].remove(ticker)
            except:
                None

    df_final = pd.DataFrame(columns=['tickers'])

    for key in df_dict.keys():
        df_final.at[key, 'tickers'] = ','.join(df_dict[key])

    df_final = df_final.sort_index()
    saved_filename = 'sp500_20200801_v2.csv'
    df_final.to_csv(HERE / saved_filename)

    print('Updated metadata file has been save as {}'.format(saved_filename))
    logging.info('Updated metadata file has been save as {}'.format(saved_filename))

    return None


@check_datasetdir_exists
@check_logdir_exists
def download_tickers_fundamental() -> pd.DataFrame:
    """
    A function to download cross-sectional fundamentals of 
    the Standard & Poor 500 listed companies based on the updated sp500 members. 
    """
    # setting logger 
    logging.basicConfig(filename= settings.logdir / 'fundamental_download.txt', level=logging.INFO,
                    format='%(asctime)s - %(levelname)s - %(message)s')

    # 1. check presence of working directory and metadata file
    working_dir = settings.datasetdir / 'metadata'
    if not os.path.exists(working_dir):
        print('metadata directory and sp500_metadata.csv not found. \
              Please initialize first by quanp.datasets.get_wiki_sp500_metadata()')
        exit(1)

    # 2. the input csv directory
    csv_dir = working_dir / 'sp500_metadata.csv'

    # the daily eod price
    try:
        df_meta = pd.read_csv(csv_dir, index_col=0)
        tickers = list(df_meta.index)
        print('Total tickers to download:', len(tickers))

    except KeyError as ke:
        print(KeyError, 'Please provide a csv that contain a column', 
                        'named {}'.format(ke))
        logging.info('Please provide a csv that contain a column \
                        named {}'.format(ke))
        exit(1)

    # create list for download failed tickers
    tickers_failed = []

    # create empty dataframe for downloaded tickers
    df_fundamental = pd.DataFrame()

    # loop and process through each provided ticker
    for ticker in tqdm(tickers, desc='Downloading fundamentals...', unit='ticker', position=0):

        # define payload
        payload = {'apikey': TOS_API_KEY,
                    'symbol': ticker,
                    'projection': 'fundamental'}

        # define the endpoint
        endpoint = "https://api.tdameritrade.com/v1/instruments"

        # request and process data
        content = requests.get(url = endpoint, params = payload)
        data = content.json()

        try:
            # dataframing and appending data
            df_tmp = pd.DataFrame.from_dict(data[ticker])
            df_tmp_fundamental = df_tmp[['fundamental']]
            df_tmp_fundamental.columns = [ticker]
            df_fundamental[ticker] = df_tmp_fundamental[ticker]
                   
        except KeyError as ke:
            print(KeyError, 'column {} not found in the processed dataframe.'.format(ke), 
                'Possible reason may be the ticker provided not found in TOS database.')
            tickers_failed.append(ticker)

    logging.info('column {} not found in the processed dataframe. Possible \
                reason may be the ticker provided not found in TOS \
                database.'.format(tickers_failed))

    df_meta_merged = pd.merge(df_meta, df_fundamental.T, how='inner', 
                                                         left_index=True,
                                                         right_index=True)

    saved_filename = 'sp500_metadata_fundamentalAdded.csv'
    df_meta_merged.to_csv(working_dir / saved_filename)

    print('Updated metadata file has been save as {}'.format(saved_filename))
    logging.info('Updated metadata file has been save as {}'.format(saved_filename))

    return df_meta_merged

@check_datasetdir_exists
@check_logdir_exists
def get_wiki_sp500_metadata() -> pd.DataFrame:
    """\
    A function to download Standard & Poor 500 listed companies according to wikipedia
    @ https://en.wikipedia.org/wiki/List_of_S%26P_500_companies
    """
    verbosity_save = settings.verbosity
    settings.verbosity = 'error'  # suppress output...
    settings.verbosity = verbosity_save

    # create output directory if not exist
    output_dir = settings.datasetdir / 'metadata'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # get html file
    htmlFile = requests.get("https://en.wikipedia.org/wiki/List_of_S%26P_500_companies")

    # parsing html file to beautisoup object
    soup = BeautifulSoup(htmlFile.text, 'lxml')

    # get table content
    data = []
    table = soup.find('table', attrs={'id': 'constituents'})
    table_body = table.find('tbody')

    rows = table_body.find_all('tr')
    for row in rows:
        cols = row.find_all('td')
        cols = [ele.text.strip() for ele in cols]
        if cols:
            data.append([ele for ele in cols if ele]) # Get rid of empty values

    # list column headers
    tableHeader = [ele.text.strip() for ele in table_body.find_all('th')]

    # Dataframing
    saved_filename = 'sp500_metadata.csv'
    df = pd.DataFrame(data, columns=tableHeader)
    df.set_index('Symbol', inplace=True, drop=True)
    df.to_csv(output_dir / saved_filename)

    print('The metadata file has been initialized and saved as {}'.format(saved_filename))
    logging.info('The metadata file has been initialized and saved as {}'.format(saved_filename))

    return df



from typing import Optional, Union, Mapping  # Special
from typing import Sequence  # ABCs
from typing import Tuple  # Classes

import numpy as np
import pandas as pd
from anndata import AnnData
from matplotlib.axes import Axes
from matplotlib import pyplot as pl

from .. import logging as logg
from .._utils import _doc_params
from .._compat import Literal
from ._utils import make_grid_spec, fix_kwds
from ._utils import ColorLike, _AxesSubplot
from ._utils import savefig_or_show
from .._settings import settings

from ._docs import doc_common_plot_args, doc_show_save_ax
from ._baseplot_class import BasePlot, doc_common_groupby_plot_args, _VarNames


@_doc_params(common_plot_args=doc_common_plot_args)
class DotPlot(BasePlot):
    """\
    Allows the visualization of two values that are encoded as
    dot size and color. The size usually represents the fraction
    of subjects (obs) that have a non-zero value for features (var).

    For each var_name and each `groupby` category a dot is plotted.
    Each dot represents two values: mean expression within each category
    (visualized by color) and fraction of subjects expressing the `var_name` in the
    category (visualized by the size of the dot). If `groupby` is not given,
    the dotplot assumes that all data belongs to a single category.

    .. note::
       A feature is considered expressed if the expression value in the `adata` (or
       `adata.raw`) is above the specified threshold which is zero by default.

    An example of dotplot usage is to visualize, for multiple marker features,
    the mean value and the percentage of subjects expressing the feature
    across multiple clusters.

    Parameters
    ----------
    {common_plot_args}
    title
        Title for the figure
    expression_cutoff
        Expression cutoff that is used for binarizing the feature expression and
        determining the fraction of subjects expressing given features. A feature is
        expressed only if the expression value is greater than this threshold.
    mean_only_expressed
        If True, feature expression is averaged only over the subjects
        expressing the given features.
    standard_scale
        Whether or not to standardize that dimension between 0 and 1,
        meaning for each variable or group,
        subtract the minimum and divide each by its maximum.
    kwds
        Are passed to :func:`matplotlib.pyplot.scatter`.

    See also
    --------
    :func:`~quanp.pl.dotplot`: Simpler way to call DotPlot but with less options.
    :func:`~quanp.pl.rank_features_groups_dotplot`: to plot marker
        features identified using the :func:`~quanp.tl.rank_features_groups` function.

    Examples
    --------

    >>> import quanp as sc
    >>> adata = qp.datasets.pbmc68k_reduced()
    >>> markers = ['featureA', 'featureB', 'featureC', 'featureD', 'featureE', 'featureF']
    >>> qp.pl.DotPlot(adata, markers, groupby='bulk_labels').show()

    Using var_names as dict:

    >>> markers = {{'A-subject': 'featureA', 'B-subject': 'featureB', 'C-subject': 'featureC'}}
    >>> qp.pl.DotPlot(adata, markers, groupby='bulk_labels').show()

    """

    DEFAULT_SAVE_PREFIX = 'dotplot_'
    # default style parameters
    DEFAULT_COLORMAP = 'winter'
    DEFAULT_COLOR_ON = 'dot'
    DEFAULT_DOT_MAX = None
    DEFAULT_DOT_MIN = None
    DEFAULT_SMALLEST_DOT = 0.0
    DEFAULT_LARGEST_DOT = 200.0
    DEFAULT_DOT_EDGECOLOR = 'black'
    DEFAULT_DOT_EDGELW = 0.2
    DEFAULT_SIZE_EXPONENT = 1.5

    # default legend parameters
    DEFAULT_SIZE_LEGEND_TITLE = 'Fraction of subjects\nin group (%)'
    DEFAULT_COLOR_LEGEND_TITLE = 'Mean expression\nin group'
    DEFAULT_LEGENDS_WIDTH = 1.5  # inches
    DEFAULT_PLOT_X_PADDING = 0.8  # a unit is the distance between two x-axis ticks
    DEFAULT_PLOT_Y_PADDING = 1.0  # a unit is the distance between two y-axis ticks

    def __init__(
        self,
        adata: AnnData,
        var_names: Union[_VarNames, Mapping[str, _VarNames]],
        groupby: Union[str, Sequence[str]],
        use_raw: Optional[bool] = None,
        log: bool = False,
        num_categories: int = 7,
        categories_order: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        feature_symbols: Optional[str] = None,
        var_group_positions: Optional[Sequence[Tuple[int, int]]] = None,
        var_group_labels: Optional[Sequence[str]] = None,
        var_group_rotation: Optional[float] = None,
        layer: Optional[str] = None,
        expression_cutoff: float = 0.0,
        mean_only_expressed: bool = False,
        standard_scale: Literal['var', 'group'] = None,
        dot_color_df: Optional[pd.DataFrame] = None,
        dot_size_df: Optional[pd.DataFrame] = None,
        ax: Optional[_AxesSubplot] = None,
        **kwds,
    ):
        BasePlot.__init__(
            self,
            adata,
            var_names,
            groupby,
            use_raw=use_raw,
            log=log,
            num_categories=num_categories,
            categories_order=categories_order,
            title=title,
            figsize=figsize,
            feature_symbols=feature_symbols,
            var_group_positions=var_group_positions,
            var_group_labels=var_group_labels,
            var_group_rotation=var_group_rotation,
            layer=layer,
            ax=ax,
            **kwds,
        )

        # for if category defined by groupby (if any) compute for each var_name
        # 1. the fraction of subjects in the category having a value >expression_cutoff
        # 2. the mean value over the category

        # 1. compute fraction of subjects having value > expression_cutoff
        # transform obs_tidy into boolean matrix using the expression_cutoff
        obs_bool = self.obs_tidy > expression_cutoff

        # compute the sum per group which in the boolean matrix this is the number
        # of values >expression_cutoff, and divide the result by the total number of
        # values in the group (given by `count()`)
        if dot_size_df is None:
            dot_size_df = (
                obs_bool.groupby(level=0).sum() / obs_bool.groupby(level=0).count()
            )

        if dot_color_df is None:
            # 2. compute mean expression value value
            if mean_only_expressed:
                dot_color_df = (
                    self.obs_tidy.mask(~obs_bool).groupby(level=0).mean().fillna(0)
                )
            else:
                dot_color_df = self.obs_tidy.groupby(level=0).mean()

            if standard_scale == 'group':
                dot_color_df = dot_color_df.sub(dot_color_df.min(1), axis=0)
                dot_color_df = dot_color_df.div(dot_color_df.max(1), axis=0).fillna(0)
            elif standard_scale == 'var':
                dot_color_df -= dot_color_df.min(0)
                dot_color_df = (dot_color_df / dot_color_df.max(0)).fillna(0)
            elif standard_scale is None:
                pass
            else:
                logg.warning('Unknown type for standard_scale, ignored')
        else:
            # check that both matrices have the same shape
            if dot_color_df.shape != dot_size_df.shape:
                logg.error(
                    "the given dot_color_df data frame has a different shape than"
                    "the data frame used for the dot size. Both data frames need"
                    "to have the same index and columns"
                )

            # Because features (columns) can be duplicated (e.g. when the
            # same feature is reported as marker feature in two clusters)
            # they need to be removed first,
            # otherwise, the duplicated features are further duplicated when reordering
            # Eg. A df with columns ['a', 'b', 'a'] after reordering columns
            # with df[['a', 'a', 'b']], results in a df with columns:
            # ['a', 'a', 'a', 'a', 'b']

            unique_var_names, unique_idx = np.unique(
                dot_color_df.columns, return_index=True
            )
            # remove duplicate columns
            if len(unique_var_names) != len(self.var_names):
                dot_color_df = dot_color_df.iloc[:, unique_idx]

            # get the same order for rows and columns in the dot_color_df
            # using the order from the doc_size_df
            dot_color_df = dot_color_df.loc[dot_size_df.index][dot_size_df.columns]

        self.dot_color_df = dot_color_df
        self.dot_size_df = dot_size_df

        # Set default style parameters
        self.cmap = self.DEFAULT_COLORMAP
        self.dot_max = self.DEFAULT_DOT_MAX
        self.dot_min = self.DEFAULT_DOT_MIN
        self.smallest_dot = self.DEFAULT_SMALLEST_DOT
        self.largest_dot = self.DEFAULT_LARGEST_DOT
        self.color_on = self.DEFAULT_COLOR_ON
        self.size_exponent = self.DEFAULT_SIZE_EXPONENT
        self.grid = False
        self.plot_x_padding = self.DEFAULT_PLOT_X_PADDING
        self.plot_y_padding = self.DEFAULT_PLOT_Y_PADDING

        self.dot_edge_color = self.DEFAULT_DOT_EDGECOLOR
        self.dot_edge_lw = self.DEFAULT_DOT_EDGELW

        # set legend defaults
        self.color_legend_title = self.DEFAULT_COLOR_LEGEND_TITLE
        self.size_title = self.DEFAULT_SIZE_LEGEND_TITLE
        self.legends_width = self.DEFAULT_LEGENDS_WIDTH
        self.show_size_legend = True
        self.show_colorbar = True

    def style(
        self,
        cmap: str = DEFAULT_COLORMAP,
        color_on: Optional[Literal['dot', 'square']] = DEFAULT_COLOR_ON,
        dot_max: Optional[float] = DEFAULT_DOT_MAX,
        dot_min: Optional[float] = DEFAULT_DOT_MIN,
        smallest_dot: Optional[float] = DEFAULT_SMALLEST_DOT,
        largest_dot: Optional[float] = DEFAULT_LARGEST_DOT,
        dot_edge_color: Optional[ColorLike] = DEFAULT_DOT_EDGECOLOR,
        dot_edge_lw: Optional[float] = DEFAULT_DOT_EDGELW,
        size_exponent: Optional[float] = DEFAULT_SIZE_EXPONENT,
        grid: Optional[float] = False,
        x_padding: Optional[float] = DEFAULT_PLOT_X_PADDING,
        y_padding: Optional[float] = DEFAULT_PLOT_Y_PADDING,
    ):
        """\
        Modifies plot visual parameters

        Parameters
        ----------
        cmap
            String denoting matplotlib color map.
        color_on
            Options are 'dot' or 'square'. Be default the colomap is applied to
            the color of the dot. Optionally, the colormap can be applied to an
            square behind the dot, in which case the dot is transparent and only
            the edge is shown.
        dot_max
            If none, the maximum dot size is set to the maximum fraction value found
            (e.g. 0.6). If given, the value should be a number between 0 and 1.
            All fractions larger than dot_max are clipped to this value.
        dot_min
            If none, the minimum dot size is set to 0. If given,
            the value should be a number between 0 and 1.
            All fractions smaller than dot_min are clipped to this value.
        smallest_dot
            If none, the smallest dot has size 0.
            All expression fractions with `dot_min` are plotted with this size.
        largest_dot
            If none, the largest dot has size 200.
            All expression fractions with `dot_max` are plotted with this size.
        dot_edge_color
            Dot edge color. When `color_on='dot'` the default is no edge. When
            `color_on='square'`, edge color is white for darker colors and black
            for lighter background square colors.
        dot_edge_lw
            Dot edge line width. When `color_on='dot'` the default is no edge. When
            `color_on='square'`, line width = 1.5.
        size_exponent
            Dot size is computed as:
            fraction  ** size exponent and afterwards scaled to match the
            `smallest_dot` and `largest_dot` size parameters.
            Using a different size exponent changes the relative sizes of the dots
            to each other.
        grid
            Set to true to show grid lines. By default grid lines are not shown.
            Further configuration of the grid lines can be achieved directly on the
            returned ax.
        x_padding
            Space between the plot left/right borders and the dots center. A unit
            is the distance between the x ticks. Only applied when color_on = dot
        y_padding
            Space between the plot top/bottom borders and the dots center. A unit is
            the distance between the y ticks. Only applied when color_on = dot

        Returns
        -------
        :class:`~quanp.pl.DotPlot`

        Examples
        -------

        >>> adata = qp.datasets.pbmc68k_reduced()
        >>> markers = ['featureA', 'featureB', 'featureC', 'featureD', 'featureE', 'featureF']

        Change color map and apply it to the square behind the dot

        >>> qp.pl.DotPlot(adata, markers, groupby='bulk_labels')\
        ...               .style(cmap='RdBu_r', color_on='square').show()

        Add edge to dots and plot a grid

        >>> qp.pl.DotPlot(adata, markers, groupby='bulk_labels')\
        ...               .style(dot_edge_color='black', dot_edge_lw=1, grid=True)\
        ...               .show()

        """

        # change only the values that had changed
        if cmap != self.cmap:
            self.cmap = cmap
        if dot_max != self.dot_max:
            self.dot_max = dot_max
        if dot_min != self.dot_min:
            self.dot_min = dot_min
        if dot_min != self.dot_min:
            self.smallest_dot = dot_min
        if largest_dot != self.largest_dot:
            self.largest_dot = largest_dot
        if color_on != self.color_on:
            self.color_on = color_on
        if size_exponent != self.size_exponent:
            self.size_exponent = size_exponent
        if dot_edge_color != self.dot_edge_color:
            self.dot_edge_color = dot_edge_color
        if dot_edge_lw != self.dot_edge_lw:
            self.dot_edge_lw = dot_edge_lw
        if grid != self.grid:
            self.grid = grid
        if x_padding != self.plot_x_padding:
            self.plot_x_padding = x_padding
        if y_padding != self.plot_y_padding:
            self.plot_y_padding = y_padding

        return self

    def legend(
        self,
        show: Optional[bool] = True,
        show_size_legend: Optional[bool] = True,
        show_colorbar: Optional[bool] = True,
        size_title: Optional[str] = DEFAULT_SIZE_LEGEND_TITLE,
        colorbar_title: Optional[str] = DEFAULT_COLOR_LEGEND_TITLE,
        width: Optional[float] = DEFAULT_LEGENDS_WIDTH,
    ):
        """\
        Configures dot size and the colorbar legends

        Parameters
        ----------
        show
            Set to `False` to hide the default plot of the legends. This sets the
            legend width to zero, which will result in a wider main plot.
        show_size_legend
            Set to `False` to hide the dot size legend
        show_colorbar
            Set to `False` to hide the colorbar legend
        size_title
            Title for the dot size legend. Use '\\n' to add line breaks. Appears on top
            of dot sizes
        colorbar_title
            Title for the color bar. Use '\\n' to add line breaks. Appears on top of the
            color bar
        width
            Width of the legends area. The unit is the same as in matplotlib (inches).

        Returns
        -------
        :class:`~quanp.pl.DotPlot`

        Examples
        --------

        Set color bar title:

        >>> adata = qp.datasets.pbmc68k_reduced()
        >>> markers = {{'T-subject': 'CD3D', 'B-subject': 'CD79A', 'myeloid': 'CST3'}}
        >>> dp = qp.pl.DotPlot(adata, markers, groupby='bulk_labels')
        >>> dp.legend(colorbar_title='log(UMI counts + 1)').show()
        """

        if not show:
            # turn of legends by setting width to 0
            self.legends_width = 0
        else:
            self.color_legend_title = colorbar_title
            self.size_title = size_title
            self.legends_width = width
            self.show_size_legend = show_size_legend
            self.show_colorbar = show_colorbar

        return self

    def _plot_size_legend(self, size_legend_ax: Axes):
        # for the dot size legend, use step between dot_max and dot_min
        # based on how different they are.
        diff = self.dot_max - self.dot_min
        if 0.3 < diff <= 0.6:
            step = 0.1
        elif diff <= 0.3:
            step = 0.05
        else:
            step = 0.2
        # a descending range that is afterwards inverted is used
        # to guarantee that dot_max is in the legend.
        size_range = np.arange(self.dot_max, self.dot_min, step * -1)[::-1]
        if self.dot_min != 0 or self.dot_max != 1:
            dot_range = self.dot_max - self.dot_min
            size_values = (size_range - self.dot_min) / dot_range
        else:
            size_values = size_range

        size = size_values ** self.size_exponent
        size = size * (self.largest_dot - self.smallest_dot) + self.smallest_dot

        # plot size bar
        size_legend_ax.scatter(
            np.arange(len(size)) + 0.5,
            np.repeat(0, len(size)),
            s=size,
            color='gray',
            edgecolor='black',
            linewidth=self.dot_edge_lw,
            zorder=100,
        )
        size_legend_ax.set_xticks(np.arange(len(size)) + 0.5)
        labels = [
            "{}".format(np.round((x * 100), decimals=0).astype(int)) for x in size_range
        ]
        size_legend_ax.set_xticklabels(labels, fontsize='small')

        # remove y ticks and labels
        size_legend_ax.tick_params(
            axis='y', left=False, labelleft=False, labelright=False
        )

        # remove surrounding lines
        size_legend_ax.spines['right'].set_visible(False)
        size_legend_ax.spines['top'].set_visible(False)
        size_legend_ax.spines['left'].set_visible(False)
        size_legend_ax.spines['bottom'].set_visible(False)
        size_legend_ax.grid(False)

        ymax = size_legend_ax.get_ylim()[1]
        size_legend_ax.set_ylim(-1.05 - self.largest_dot * 0.003, 4)
        size_legend_ax.set_title(self.size_title, y=ymax + 0.45, size='small')

        xmin, xmax = size_legend_ax.get_xlim()
        size_legend_ax.set_xlim(xmin - 0.15, xmax + 0.5)

    def _plot_legend(self, legend_ax, return_ax_dict, normalize):

        # to maintain the fixed height size of the legends, a
        # spacer of variable height is added at the bottom.
        # The structure for the legends is:
        # first row: variable space to keep the other rows of
        #            the same size (avoid stretching)
        # second row: legend for dot size
        # third row: spacer to avoid color and size legend titles to overlap
        # fourth row: colorbar

        cbar_legend_height = self.min_figure_height * 0.08
        size_legend_height = self.min_figure_height * 0.27
        spacer_height = self.min_figure_height * 0.3

        height_ratios = [
            self.height - size_legend_height - cbar_legend_height - spacer_height,
            size_legend_height,
            spacer_height,
            cbar_legend_height,
        ]
        fig, legend_gs = make_grid_spec(
            legend_ax, nrows=4, ncols=1, height_ratios=height_ratios,
        )

        if self.show_size_legend:
            size_legend_ax = fig.add_subplot(legend_gs[1])
            self._plot_size_legend(size_legend_ax)
            return_ax_dict['size_legend_ax'] = size_legend_ax

        if self.show_colorbar:
            color_legend_ax = fig.add_subplot(legend_gs[3])

            self._plot_colorbar(color_legend_ax, normalize)
            return_ax_dict['color_legend_ax'] = color_legend_ax

    def _mainplot(self, ax):
        # work on a copy of the dataframes. This is to avoid changes
        # on the original data frames after repetitive calls to the
        # DotPlot object, for example once with swap_axes and other without

        _color_df = self.dot_color_df.copy()
        _size_df = self.dot_size_df.copy()
        if self.var_names_idx_order is not None:
            _color_df = _color_df.iloc[:, self.var_names_idx_order]
            _size_df = _size_df.iloc[:, self.var_names_idx_order]

        if self.categories_order is not None:
            _color_df = _color_df.loc[self.categories_order, :]
            _size_df = _size_df.loc[self.categories_order, :]

        if self.are_axes_swapped:
            _size_df = _size_df.T
            _color_df = _color_df.T
        self.cmap = self.kwds.get('cmap', self.cmap)
        if 'cmap' in self.kwds:
            del self.kwds['cmap']

        normalize, dot_min, dot_max = self._dotplot(
            _size_df,
            _color_df,
            ax,
            cmap=self.cmap,
            dot_max=self.dot_max,
            dot_min=self.dot_min,
            color_on=self.color_on,
            edge_color=self.dot_edge_color,
            edge_lw=self.dot_edge_lw,
            smallest_dot=self.smallest_dot,
            largest_dot=self.largest_dot,
            size_exponent=self.size_exponent,
            grid=self.grid,
            x_padding=self.plot_x_padding,
            y_padding=self.plot_y_padding,
            **self.kwds,
        )

        self.dot_min, self.dot_max = dot_min, dot_max
        return normalize

    @staticmethod
    def _dotplot(
        dot_size,
        dot_color,
        dot_ax,
        cmap: str = 'Reds',
        color_on: Optional[str] = 'dot',
        y_label: Optional[str] = None,
        dot_max: Optional[float] = None,
        dot_min: Optional[float] = None,
        standard_scale: Literal['var', 'group'] = None,
        smallest_dot: Optional[float] = 0.0,
        largest_dot: Optional[float] = 200,
        size_exponent: Optional[float] = 2,
        edge_color: Optional[ColorLike] = None,
        edge_lw: Optional[float] = None,
        grid: Optional[bool] = False,
        x_padding: Optional[float] = 0.8,
        y_padding: Optional[float] = 1.0,
        **kwds,
    ):
        """\
        Makes a *dot plot* given two data frames, one containing
        the doc size and other containing the dot color. The indices and
        columns of the data frame are used to label the output image

        The dots are plotted using :func:`matplotlib.pyplot.scatter`. Thus, additional
        arguments can be passed.

        Parameters
        ----------
        dot_size: Data frame containing the dot_size.
        dot_color: Data frame containing the dot_color, should have the same,
                shape, columns and indices as dot_size.
        dot_ax: matplotlib axis
        cmap
            String denoting matplotlib color map.
        color_on
            Options are 'dot' or 'square'. Be default the colomap is applied to
            the color of the dot. Optionally, the colormap can be applied to an
            square behind the dot, in which case the dot is transparent and only
            the edge is shown.
        y_label: String. Label for y axis
        dot_max
            If none, the maximum dot size is set to the maximum fraction value found
            (e.g. 0.6). If given, the value should be a number between 0 and 1.
            All fractions larger than dot_max are clipped to this value.
        dot_min
            If none, the minimum dot size is set to 0. If given,
            the value should be a number between 0 and 1.
            All fractions smaller than dot_min are clipped to this value.
        standard_scale
            Whether or not to standardize that dimension between 0 and 1,
            meaning for each variable or group,
            subtract the minimum and divide each by its maximum.
        smallest_dot
            If none, the smallest dot has size 0.
            All expression levels with `dot_min` are plotted with this size.
        edge_color
            Dot edge color. When `color_on='dot'` the default is no edge. When
            `color_on='square'`, edge color is white
        edge_lw
            Dot edge line width. When `color_on='dot'` the default is no edge. When
            `color_on='square'`, line width = 1.5
        grid
            Adds a grid to the plot
        x_paddding
            Space between the plot left/right borders and the dots center. A unit
            is the distance between the x ticks. Only applied when color_on = dot
        y_paddding
            Space between the plot top/bottom borders and the dots center. A unit is
            the distance between the y ticks. Only applied when color_on = dot
        kwds
            Are passed to :func:`matplotlib.pyplot.scatter`.

        Returns
        -------
        matplotlib.colors.Normalize, dot_min, dot_max

        """
        assert dot_size.shape == dot_color.shape, (
            'please check that dot_size ' 'and dot_color dataframes have the same shape'
        )

        assert list(dot_size.index) == list(dot_color.index), (
            'please check that dot_size ' 'and dot_color dataframes have the same index'
        )

        assert list(dot_size.columns) == list(dot_color.columns), (
            'please check that the dot_size '
            'and dot_color dataframes have the same columns'
        )

        if standard_scale == 'group':
            dot_color = dot_color.sub(dot_color.min(1), axis=0)
            dot_color = dot_color.div(dot_color.max(1), axis=0).fillna(0)
        elif standard_scale == 'var':
            dot_color -= dot_color.min(0)
            dot_color = (dot_color / dot_color.max(0)).fillna(0)
        elif standard_scale is None:
            pass

        # make scatter plot in which
        # x = var_names
        # y = groupby category
        # size = fraction
        # color = mean expression

        # +0.5 in y and x to set the dot center at 0.5 multiples
        # this facilitates dendrogram and totals alignment for
        # matrixplot, dotplot and stackec_violin using the same coordinates.
        y, x = np.indices(dot_color.shape)
        y = y.flatten() + 0.5
        x = x.flatten() + 0.5
        frac = dot_size.values.flatten()
        mean_flat = dot_color.values.flatten()
        cmap = pl.get_cmap(kwds.get('cmap', cmap))
        if 'cmap' in kwds:
            del kwds['cmap']
        if dot_max is None:
            dot_max = np.ceil(max(frac) * 10) / 10
        else:
            if dot_max < 0 or dot_max > 1:
                raise ValueError("`dot_max` value has to be between 0 and 1")
        if dot_min is None:
            dot_min = 0
        else:
            if dot_min < 0 or dot_min > 1:
                raise ValueError("`dot_min` value has to be between 0 and 1")

        if dot_min != 0 or dot_max != 1:
            # clip frac between dot_min and  dot_max
            frac = np.clip(frac, dot_min, dot_max)
            old_range = dot_max - dot_min
            # re-scale frac between 0 and 1
            frac = (frac - dot_min) / old_range

        size = frac ** size_exponent
        # rescale size to match smallest_dot and largest_dot
        size = size * (largest_dot - smallest_dot) + smallest_dot

        import matplotlib.colors

        normalize = matplotlib.colors.Normalize(
            vmin=kwds.get('vmin'), vmax=kwds.get('vmax')
        )

        if color_on == 'square':
            if edge_color is None:
                from seaborn.utils import relative_luminance

                # use either black or white for the edge color
                # depending on the luminance of the background
                # square color
                edge_color = []
                for color_value in cmap(normalize(mean_flat)):
                    lum = relative_luminance(color_value)
                    edge_color.append(".15" if lum > 0.408 else "w")

            edge_lw = 1.5 if edge_lw is None else edge_lw

            # first make a heatmap similar to `qp.pl.matrixplot`
            # (squares with the asigned colormap). Circles will be plotted
            # on top
            dot_ax.pcolor(dot_color.values, cmap=cmap, norm=normalize)
            for axis in ['top', 'bottom', 'left', 'right']:
                dot_ax.spines[axis].set_linewidth(1.5)
            kwds = fix_kwds(
                kwds,
                s=size,
                cmap=cmap,
                norm=None,
                linewidth=edge_lw,
                facecolor='none',
                edgecolor=edge_color,
            )
            dot_ax.scatter(x, y, **kwds)
        else:
            edge_color = 'none' if edge_color is None else edge_color
            edge_lw = 0.0 if edge_lw is None else edge_lw

            color = cmap(normalize(mean_flat))
            kwds = fix_kwds(
                kwds,
                s=size,
                cmap=cmap,
                color=color,
                norm=None,
                linewidth=edge_lw,
                edgecolor=edge_color,
            )

            dot_ax.scatter(x, y, **kwds)

        y_ticks = np.arange(dot_color.shape[0]) + 0.5
        dot_ax.set_yticks(y_ticks)
        dot_ax.set_yticklabels(
            [dot_color.index[idx] for idx, _ in enumerate(y_ticks)], minor=False
        )

        x_ticks = np.arange(dot_color.shape[1]) + 0.5
        dot_ax.set_xticks(x_ticks)
        dot_ax.set_xticklabels(
            [dot_color.columns[idx] for idx, _ in enumerate(x_ticks)],
            rotation=90,
            ha='center',
            minor=False,
        )
        dot_ax.tick_params(axis='both', labelsize='small')
        dot_ax.grid(False)
        dot_ax.set_ylabel(y_label)

        # to be consistent with the heatmap plot, is better to
        # invert the order of the y-axis, such that the first group is on
        # top
        dot_ax.set_ylim(dot_color.shape[0], 0)
        dot_ax.set_xlim(0, dot_color.shape[1])

        if color_on == 'dot':
            # add padding to the x and y lims when the color is not in the square
            # default y range goes from 0.5 to num cols + 0.5
            # and default x range goes from 0.5 to num rows + 0.5, thus
            # the padding needs to be corrected.
            x_padding = x_padding - 0.5
            y_padding = y_padding - 0.5
            dot_ax.set_ylim(dot_color.shape[0] + y_padding, -y_padding)

            dot_ax.set_xlim(-x_padding, dot_color.shape[1] + x_padding)

        if grid:
            dot_ax.grid(True, color='gray', linewidth=0.1)
            dot_ax.set_axisbelow(True)

        return normalize, dot_min, dot_max


@_doc_params(
    show_save_ax=doc_show_save_ax,
    common_plot_args=doc_common_plot_args,
    groupby_plots_args=doc_common_groupby_plot_args,
)
def dotplot(
    adata: AnnData,
    var_names: Union[_VarNames, Mapping[str, _VarNames]],
    groupby: Union[str, Sequence[str]],
    use_raw: Optional[bool] = None,
    log: bool = False,
    num_categories: int = 7,
    expression_cutoff: float = 0.0,
    mean_only_expressed: bool = False,
    cmap: str = 'Reds',
    dot_max: Optional[float] = DotPlot.DEFAULT_DOT_MAX,
    dot_min: Optional[float] = DotPlot.DEFAULT_DOT_MIN,
    standard_scale: Optional[Literal['var', 'group']] = None,
    smallest_dot: Optional[float] = DotPlot.DEFAULT_SMALLEST_DOT,
    title: Optional[str] = None,
    colorbar_title: Optional[str] = DotPlot.DEFAULT_COLOR_LEGEND_TITLE,
    size_title: Optional[str] = DotPlot.DEFAULT_SIZE_LEGEND_TITLE,
    figsize: Optional[Tuple[float, float]] = None,
    dendrogram: Union[bool, str] = False,
    feature_symbols: Optional[str] = None,
    var_group_positions: Optional[Sequence[Tuple[int, int]]] = None,
    var_group_labels: Optional[Sequence[str]] = None,
    var_group_rotation: Optional[float] = None,
    layer: Optional[str] = None,
    swap_axes: Optional[bool] = False,
    dot_color_df: Optional[pd.DataFrame] = None,
    show: Optional[bool] = None,
    save: Union[str, bool, None] = None,
    ax: Optional[_AxesSubplot] = None,
    return_fig: Optional[bool] = False,
    **kwds,
) -> Union[DotPlot, dict, None]:
    """\
    Makes a *dot plot* of the expression values of `var_names`.

    For each var_name and each `groupby` category a dot is plotted.
    Each dot represents two values: mean expression within each category
    (visualized by color) and fraction of subjects expressing the `var_name` in the
    category (visualized by the size of the dot). If `groupby` is not given,
    the dotplot assumes that all data belongs to a single category.

    .. note::
       A feature is considered expressed if the expression value in the `adata` (or
       `adata.raw`) is above the specified threshold which is zero by default.

    An example of dotplot usage is to visualize, for multiple marker features,
    the mean value and the percentage of subjects expressing the feature
    across  multiple clusters.

    This function provides a convenient interface to the :class:`~quanp.pl.DotPlot`
    class. If you need more flexibility, you should use :class:`~quanp.pl.DotPlot`
    directly.

    Parameters
    ----------
    {common_plot_args}
    {groupby_plots_args}
    size_title
        Title for the size legend. New line character (\\n) can be used.
    expression_cutoff
        Expression cutoff that is used for binarizing the feature expression and
        determining the fraction of subjects expressing given features. A feature is
        expressed only if the expression value is greater than this threshold.
    mean_only_expressed
        If True, feature expression is averaged only over the subjects
        expressing the given features.
    dot_max
        If none, the maximum dot size is set to the maximum fraction value found
        (e.g. 0.6). If given, the value should be a number between 0 and 1.
        All fractions larger than dot_max are clipped to this value.
    dot_min
        If none, the minimum dot size is set to 0. If given,
        the value should be a number between 0 and 1.
        All fractions smaller than dot_min are clipped to this value.
    smallest_dot
        If none, the smallest dot has size 0.
        All expression levels with `dot_min` are plotted with this size.
    {show_save_ax}
    kwds
        Are passed to :func:`matplotlib.pyplot.scatter`.

    Returns
    -------
    If `return_fig` is `True`, returns a :class:`~quanp.pl.DotPlot` object,
    else if `show` is false, return axes dict

    See also
    --------
    :class:`~quanp.pl.DotPlot`: The DotPlot class can be used to to control
        several visual parameters not available in this function.
    :func:`~quanp.pl.rank_features_groups_dotplot`: to plot marker features
        identified using the :func:`~quanp.tl.rank_features_groups` function.

    Examples
    --------

    Create a dot plot using the given markers and the PBMC example dataset grouped by
    the category 'bulk_labels'.

    >>> import quanp as qp
    >>> adata = qp.datasets.pbmc68k_reduced()
    >>> markers = ['featureA', 'featureB', 'featureC', 'featureD', 'featureE', 'featureF']
    >>> qp.pl.dotplot(adata, markers, groupby='bulk_labels', dendrogram=True)

    Using var_names as dict:

    >>> markers = {{'A-subject': 'featureA', 'B-subject': 'featureB', 'C-subject': 'featureC'}}
    >>> qp.pl.dotplot(adata, markers, groupby='bulk_labels', dendrogram=True)

    Get DotPlot object for fine tuning

    >>> dp = qp.pl.dotplot(adata, markers, 'bulk_labels', return_fig=True)
    >>> dp.add_totals().style(dot_edge_color='black', dot_edge_lw=0.5).show()

    The axes used can be obtained using the get_axes() method

    >>> axes_dict = dp.get_axes()
    >>> print(axes_dict)

    """

    # backwards compatibility: previous version of dotplot used `color_map`
    # instead of `cmap`
    cmap = kwds.get('color_map', cmap)
    if 'color_map' in kwds:
        del kwds['color_map']

    dp = DotPlot(
        adata,
        var_names,
        groupby,
        use_raw=use_raw,
        log=log,
        num_categories=num_categories,
        expression_cutoff=expression_cutoff,
        mean_only_expressed=mean_only_expressed,
        standard_scale=standard_scale,
        title=title,
        figsize=figsize,
        feature_symbols=feature_symbols,
        var_group_positions=var_group_positions,
        var_group_labels=var_group_labels,
        var_group_rotation=var_group_rotation,
        layer=layer,
        dot_color_df=dot_color_df,
        ax=ax,
        **kwds,
    )

    if dendrogram:
        dp.add_dendrogram(dendrogram_key=dendrogram)
    if swap_axes:
        dp.swap_axes()

    dp = dp.style(
        cmap=cmap,
        dot_max=dot_max,
        dot_min=dot_min,
        smallest_dot=smallest_dot,
        dot_edge_lw=kwds.pop('linewidth', DotPlot.DEFAULT_DOT_EDGELW),
    ).legend(colorbar_title=colorbar_title, size_title=size_title,)

    if return_fig:
        return dp
    else:
        dp.make_figure()
        savefig_or_show(DotPlot.DEFAULT_SAVE_PREFIX, show=show, save=save)
        show = settings.autoshow if show is None else show
        if not show:
            return dp.get_axes()



# Author: Alex Wolf (http://falexwolf.de)
"""Simulate Data

Simulate stochastic dynamic systems to model feature expression dynamics and
cause-effect data.

TODO
----
Beta Version. The code will be reorganized soon.
"""

import itertools
import shutil
import sys
from pathlib import Path
from types import MappingProxyType
from typing import Optional, Union, List, Tuple, Mapping

import numpy as np
import scipy as sp
from anndata import AnnData

from .. import _utils, readwrite, logging as logg
from .._settings import settings
from .._compat import Literal


def sim(
    model: Literal['krumsiek11', 'toggleswitch'],
    params_file: bool = True,
    tmax: Optional[int] = None,
    branching: Optional[bool] = None,
    nrRealizations: Optional[int] = None,
    noiseObs: Optional[float] = None,
    noiseDyn: Optional[float] = None,
    step: Optional[int] = None,
    seed: Optional[int] = None,
    writedir: Optional[Union[str, Path]] = None,
) -> AnnData:
    """\
    Simulate dynamic feature expression data [Wittmann09]_ [Wolf18]_.

    Sample from a stochastic differential equation model built from
    literature-curated boolean feature regulatory networks, as suggested by
    [Wittmann09]_. The Quanpy implementation is due to [Wolf18]_.

    Parameters
    ----------
    model
        Model file in 'sim_models' directory.
    params_file
        Read default params from file.
    tmax
        Number of time steps per realization of time series.
    branching
        Only write realizations that contain new branches.
    nrRealizations
        Number of realizations.
    noiseObs
        Observatory/Measurement noise.
    noiseDyn
        Dynamic noise.
    step
        Interval for saving state of system.
    seed
        Seed for generation of random numbers.
    writedir
        Path to directory for writing output files.

    Returns
    -------
    Annotated data matrix.

    Examples
    --------
    See this `use case <https://github.com/theislab/scanpy_usage/tree/master/170430_krumsiek11>`__
    """
    params = locals()
    if params_file:
        model_key = Path(model).with_suffix('').name
        from .. import sim_models
        pfile_sim = (
            Path(sim_models.__file__).parent
            / f'{model_key}_params.txt'
        )
        default_params = readwrite.read_params(pfile_sim)
        params = _utils.update_params(default_params, params)
    adata = sample_dynamic_data(**params)
    adata.uns['iroot'] = 0
    return adata


def add_args(p):
    """
    Update parser with tool specific arguments.

    This overwrites was is done in utils.uns_args.
    """
    # dictionary for adding arguments
    dadd_args = {
        '--opfile': {
            'default': '',
            'metavar': 'f',
            'type': str,
            'help': 'Specify a parameter file '
                    '(default: "sim/${exkey}_params.txt")'}}
    p = _utils.add_args(p, dadd_args)
    return p


def sample_dynamic_data(**params):
    """
    Helper function.
    """
    model_key = Path(params['model']).with_suffix('').name
    writedir = params.get('writedir')
    if writedir is None:
        writedir = settings.writedir / (model_key + '_sim')
    else:
        writedir = Path(writedir)
    writedir.mkdir(parents=True, exist_ok=True)
    readwrite.write_params(writedir / 'params.txt', params)
    # init variables
    tmax = params['tmax']
    branching = params['branching']
    noiseObs = params['noiseObs']
    noiseDyn = params['noiseDyn']
    nrRealizations = params['nrRealizations']
    step = params['step']  # step size for saving the figure

    nrSamples = 1  # how many files?
    maxRestarts = 1000
    maxNrSamples = 1

    # simple vector auto regressive process or
    # hill kinetics process simulation
    if 'krumsiek11' not in model_key:
        # create instance, set seed
        grnsim = GRNsim(model=model_key, params=params)
        nrOffEdges_list = np.zeros(nrSamples)
        for sample in range(nrSamples):
            # random topology / for a given edge density
            if 'hill' not in model_key:
                Coupl = np.array(grnsim.Coupl)
                for sampleCoupl in range(10):
                    nrOffEdges = 0
                    for gp in range(grnsim.dim):
                        for g in range(grnsim.dim):
                            # only consider off-diagonal edges
                            if g != gp:
                                Coupl[gp, g] = 0.7 if np.random.rand() < 0.4 else 0
                                nrOffEdges += 1 if Coupl[gp, g] > 0 else 0
                            else:
                                Coupl[gp, g] = 0.7
                    # check that the coupling matrix does not have eigenvalues
                    # greater than 1, which would lead to an exploding var process
                    if max(sp.linalg.eig(Coupl)[0]) < 1:
                        break
                nrOffEdges_list[sample] = nrOffEdges
                grnsim.set_coupl(Coupl)
            # init type
            real = 0
            X0 = np.random.rand(grnsim.dim)
            Xsamples = []
            for restart in range(nrRealizations+maxRestarts):
                # slightly break symmetry in initial conditions
                if 'toggleswitch' in model_key:
                    X0 = (np.array([0.8 for i in range(grnsim.dim)])
                          + 0.01*np.random.randn(grnsim.dim))
                X = grnsim.sim_model(tmax=tmax, X0=X0,
                                     noiseDyn=noiseDyn)
                # check branching
                check = True
                if branching:
                    check, Xsamples = _check_branching(X, Xsamples, restart)
                if check:
                    real += 1
                    grnsim.write_data(X[::step], dir=writedir,
                                      noiseObs=noiseObs,
                                      append=(False if restart==0 else True),
                                      branching=branching,
                                      nrRealizations=nrRealizations)
                # append some zeros
                if 'zeros' in writedir.name and real == 2:
                    grnsim.write_data(noiseDyn*np.random.randn(500,3), dir=writedir,
                                      noiseObs=noiseObs,
                                      append=(False if restart==0 else True),
                                      branching=branching,
                                      nrRealizations=nrRealizations)
                if real >= nrRealizations:
                    break
        logg.debug(
            f'mean nr of offdiagonal edges {nrOffEdges_list.mean()} '
            f'compared to total nr {grnsim.dim*(grnsim.dim-1)/2.}'
        )

    # more complex models
    else:
        initType = 'random'

        dim = 11
        step = 5

        grnsim = GRNsim(dim=dim,initType=initType,model=model_key,params=params)
        curr_nrSamples = 0
        Xsamples = []
        for sample in range(maxNrSamples):
            # choose initial conditions such that branchings result
            if initType == 'branch':
                X0mean = grnsim.branch_init_model1(tmax)
                if X0mean is None:
                    grnsim.set_coupl()
                    continue
            real = 0
            for restart in range(nrRealizations+maxRestarts):
                if initType == 'branch':
                    # vary initial conditions around mean
                    X0 = X0mean + (0.05*np.random.rand(dim) - 0.025*np.ones(dim))
                else:
                    # generate random initial conditions within [0.3,0.7]
                    X0 = 0.4*np.random.rand(dim)+0.3
                if model_key in [5,6]:
                    X0 = np.array([0.3,0.3,0,0,0,0])
                if model_key in [7,8,9,10]:
                    X0 = 0.6*np.random.rand(dim)+0.2
                    X0[2:] = np.zeros(4)
                if 'krumsiek11' in model_key:
                    X0 = np.zeros(dim)
                    X0[grnsim.varNames['Gata2']] = 0.8
                    X0[grnsim.varNames['Pu.1']] = 0.8
                    X0[grnsim.varNames['Cebpa']] = 0.8
                    X0 += 0.001*np.random.randn(dim)
                    if False:
                        switch_feature = restart - (nrRealizations - dim)
                        if switch_feature >= dim:
                            break
                        X0[switch_feature] = 0 if X0[switch_feature] > 0.1 else 0.8
                X = grnsim.sim_model(tmax,X0=X0,
                                     noiseDyn=noiseDyn,
                                     restart=restart)
                # check branching
                check = True
                if branching:
                    check, Xsamples = _check_branching(X,Xsamples,restart)
                if check:
                    real += 1
                    grnsim.write_data(X[::step],dir=writedir,noiseObs=noiseObs,
                                     append=(False if restart==0 else True),
                                     branching=branching,
                                     nrRealizations=nrRealizations)
                if real >= nrRealizations:
                    break
    # load the last simulation file
    filename = None
    for filename in writedir.glob('sim*.txt'):
        pass
    logg.info(f'reading simulation results {filename}')
    adata = readwrite._read(filename, first_column_names=True,
                            suppress_cache_warning=True)
    adata.uns['tmax_write'] = tmax/step
    return adata


def write_data(
    X,
    dir=Path('sim/test'),
    append=False,
    header='',
    varNames: Mapping[str, int] = MappingProxyType({}),
    Adj=np.array([]),
    Coupl=np.array([]),
    boolRules: Mapping[str, str] = MappingProxyType({}),
    model='',
    modelType='',
    invTimeStep=1,
):
    """ Write simulated data.

        Accounts for saving at the same time an ID
        and a model file.
    """
    dir.mkdir(parents=True, exist_ok=True)
    # update file with sample ids
    filename = dir / 'id.txt'
    if filename.is_file():
        with filename.open('r') as f:
            id = int(f.read()) + (0 if append else 1)
    else:
        id = 0
    with filename.open('w') as f:
        id = '{:0>6}'.format(id)
        f.write(str(id))
    # dimension
    dim = X.shape[1]
    # write files with adjacancy and coupling matrices
    if not append:
        if False:
            if Adj.size > 0:
                # due to 'update formulation' of model, there
                # is always a diagonal dependence
                Adj = np.copy(Adj)
                if 'hill' in model:
                    for i in range(Adj.shape[0]):
                        Adj[i,i] = 1
                np.savetxt(dir+'/adj_'+id+'.txt',Adj,
                           header=header,
                           fmt='%d')
            if Coupl.size > 0:
                np.savetxt(dir+'/coupl_'+id+'.txt',Coupl,
                           header=header,
                           fmt='%10.6f')
        # write model file
        if varNames and Coupl.size > 0:
            with (dir / f'model_{id}.txt').open('w') as f:
                f.write('# For each "variable = ", there must be a right hand side: \n')
                f.write('# either an empty string or a python-style logical expression \n')
                f.write('# involving variable names, "or", "and", "(", ")". \n')
                f.write('# The order of equations matters! \n')
                f.write('# \n')
                f.write('# modelType = ' + modelType + '\n')
                f.write('# invTimeStep = '+ str(invTimeStep) + '\n')
                f.write('# \n')
                f.write('# boolean update rules: \n')
                for k, v in boolRules.items():
                    f.write(f'{k} = {v}\n')
                # write coupling via names
                f.write('# coupling list: \n')
                names = list(varNames.keys())
                for gp in range(dim):
                    for g in range(dim):
                        if np.abs(Coupl[gp, g]) > 1e-10:
                            f.write(
                                f'{names[gp]:10} '
                                f'{names[g]:10} '
                                f'{Coupl[gp, g]:10.3} \n'
                            )
    # write simulated data
    # the binary mode option in the following line is a fix for python 3
    # variable names
    if varNames:
        header += f'{"it":>2} '
        for v in varNames.keys():
            header += f'{v:>7} '
    with (dir / f'sim_{id}.txt').open('ab' if append else 'wb') as f:
        np.savetxt(
            f,
            np.c_[np.arange(0, X.shape[0]), X],
            header=('' if append else header),
            fmt=['%4.f']+['%7.4f' for i in range(dim)],
        )


class GRNsim:
    """
    Simlulation of stochastic dynamic systems.

    Main application: simulation of feature expression dynamics.

    Also standard models are implemented.
    """

    availModels = dict(
        krumsiek11=(
            'myeloid progenitor network, Krumsiek et al., PLOS One 6, e22649, '
            '\n      equations from Table 1 on page 3, '
            'doi:10.1371/journal.pone.0022649 \n'
        ),
        var='vector autoregressive process \n',
        hill='process with hill kinetics \n',
    )

    writeOutputOnce = True

    def __init__(
        self,
        dim=3,
        model='ex0',
        modelType='var',
        initType='random',
        show=False,
        verbosity=0,
        Coupl=None,
        params=MappingProxyType({}),
    ):
        """
        Params
        ------
        model
            either string for predefined model,
            or directory with a model file and a couple matrix files
        """
        self.dim = dim if Coupl is None else Coupl.shape[0]  # number of nodes / dimension of system
        self.maxnpar = 1    # maximal number of parents
        self.p_indep = 0.4  # fraction of independent features
        self.model = model
        self.modelType = modelType
        self.initType = initType  # string characterizing a specific initial
        self.show = show
        self.verbosity = verbosity
        # checks
        if initType not in ['branch','random']:
            raise RuntimeError('initType must be either: branch, random')
        read = False
        if model not in self.availModels.keys():
            message = 'model not among predefined models \n'
        # read from file
        from .. import sim_models
        model = Path(sim_models.__file__).parent / f'{model}.txt'
        if not model.is_file():
            raise RuntimeError(f'Model file {model} does not exist')
        self.model = model
        # set the coupling matrix, and with that the adjacency matrix
        self.set_coupl(Coupl=Coupl)
        # seed
        np.random.seed(params['seed'])
        # header
        self.header = 'model = ' + self.model.name + ' \n'
        # params
        self.params = params

    def sim_model(self, tmax, X0, noiseDyn=0, restart=0):
        """ Simulate the model.
        """
        self.noiseDyn = noiseDyn
        #
        X = np.zeros((tmax, self.dim))
        X[0] = X0 + noiseDyn*np.random.randn(self.dim)
        # run simulation
        for t in range(1, tmax):
            if self.modelType == 'hill':
                Xdiff = self.Xdiff_hill(X[t-1])
            elif self.modelType == 'var':
                Xdiff = self.Xdiff_var(X[t-1])
            else:
                raise ValueError(f"Unknown modelType {self.modelType!r}")
            X[t] = X[t-1] + Xdiff
            # add dynamic noise
            X[t] += noiseDyn*np.random.randn(self.dim)
        return X

    def Xdiff_hill(self,Xt):
        """ Build Xdiff from coefficients of boolean network,
            that is, using self.boolCoeff. The employed functions
            are Hill type activation and deactivation functions.

            See Wittmann et al., BMC Syst. Biol. 3, 98 (2009),
            doi:10.1186/1752-0509-3-98 for more details.
        """
        verbosity = self.verbosity>0 and self.writeOutputOnce
        self.writeOutputOnce = False
        Xdiff = np.zeros(self.dim)
        for ichild,child in enumerate(self.pas.keys()):
            # check whether list of parents is non-empty,
            # otherwise continue
            if self.pas[child]:
                Xdiff_syn = 0  # synthesize term
                if verbosity > 0:
                    Xdiff_syn_str = ''
            else:
                continue
            # loop over all tuples for which the boolean update
            # rule returns true, these are stored in self.boolCoeff
            for ituple,tuple in enumerate(self.boolCoeff[child]):
                Xdiff_syn_tuple = 1
                Xdiff_syn_tuple_str = ''
                for iv,v in enumerate(tuple):
                    iparent = self.varNames[self.pas[child][iv]]
                    x = Xt[iparent]
                    threshold = 0.1/np.abs(self.Coupl[ichild,iparent])
                    Xdiff_syn_tuple *= self.hill_a(x,threshold) if v else self.hill_i(x,threshold)
                    if verbosity > 0:
                        Xdiff_syn_tuple_str += (
                            f'{"a" if v else "i"}'
                            f'({self.pas[child][iv]}, {threshold:.2})'
                        )
                Xdiff_syn += Xdiff_syn_tuple
                if verbosity > 0:
                    Xdiff_syn_str += ('+' if ituple != 0 else '') + Xdiff_syn_tuple_str
            # multiply with degradation term
            Xdiff[ichild] = self.invTimeStep*(Xdiff_syn - Xt[ichild])
            if verbosity > 0:
                Xdiff_str = (
                    f'{child}_{child}-{child} = '
                    f'{self.invTimeStep}*({Xdiff_syn_str}-{child})'
                )
                settings.m(0, Xdiff_str)
        return Xdiff

    def Xdiff_var(self, Xt, verbosity=0):
         """
         """
         # subtract the current state
         Xdiff = -Xt
         # add the information from the past
         Xdiff += np.dot(self.Coupl,Xt)
         return Xdiff

    def hill_a(self, x, threshold=0.1, power=2):
        """ Activating hill function. """
        x_pow = np.power(x,power)
        threshold_pow = np.power(threshold,power)
        return x_pow / (x_pow + threshold_pow)

    def hill_i(self, x, threshold=0.1, power=2):
        """ Inhibiting hill function.

            Is equivalent to 1-hill_a(self,x,power,threshold).
        """
        x_pow = np.power(x, power)
        threshold_pow = np.power(threshold,power)
        return threshold_pow / (x_pow + threshold_pow)

    def nhill_a(self, x, threshold=0.1, power=2, ichild=2):
        """ Normalized activating hill function. """
        x_pow = np.power(x, power)
        threshold_pow = np.power(threshold, power)
        return x_pow / (x_pow + threshold_pow) * (1 + threshold_pow)

    def nhill_i(self, x, threshold=0.1, power=2):
        """ Normalized inhibiting hill function.

            Is equivalent to 1-nhill_a(self,x,power,threshold).
        """
        x_pow = np.power(x, power)
        threshold_pow = np.power(threshold, power)
        return threshold_pow / (x_pow + threshold_pow) * (1 - x_pow)

    def read_model(self):
        """ Read the model and the couplings from the model file.
        """
        if self.verbosity > 0:
            settings.m(0, 'reading model', self.model)
        # read model
        boolRules = []
        for line in self.model.open():
            if line.startswith('#') and 'modelType =' in line:
                keyval = line
                if '|' in line:
                    keyval, type = line.split('|')[:2]
                self.modelType = keyval.split('=')[1].strip()
            if line.startswith('#') and 'invTimeStep =' in line:
                keyval = line
                if '|' in line:
                    keyval, type = line.split('|')[:2]
                self.invTimeStep = float(keyval.split('=')[1].strip())
            if not line.startswith('#'):
                boolRules.append([s.strip() for s in line.split('=')])
            if line.startswith('# coupling list:'):
                break
        self.dim = len(boolRules)
        self.boolRules = dict(boolRules)
        self.varNames = {s: i for i, s in enumerate(self.boolRules.keys())}
        names = self.varNames
        # read couplings via names
        self.Coupl = np.zeros((self.dim, self.dim))
        boolContinue = True
        for line in self.model.open():  # open(self.model.replace('/model','/couplList')):
            if line.startswith('# coupling list:'):
                boolContinue = False
            if boolContinue:
                continue
            if not line.startswith('#'):
                gps, gs, val = line.strip().split()
                self.Coupl[int(names[gps]), int(names[gs])] = float(val)
        # adjancecy matrices
        self.Adj_signed = np.sign(self.Coupl)
        self.Adj = np.abs(np.array(self.Adj_signed))
        # build bool coefficients (necessary for odefy type
        # version of the discrete model)
        self.build_boolCoeff()

    def set_coupl(self, Coupl=None):
        """ Construct the coupling matrix (and adjacancy matrix) from predefined models
            or via sampling.
        """
        self.varNames = {str(i): i for i in range(self.dim)}
        if (self.model not in self.availModels.keys()
            and Coupl is None):
            self.read_model()
        elif 'var' in self.model.name:
            # vector auto regressive process
            self.Coupl = Coupl
            self.boolRules = {s: '' for s in self.varNames.keys()}
            names = list(self.varNames.keys())
            for gp in range(self.dim):
                pas = []
                for g in range(self.dim):
                    if np.abs(self.Coupl[gp,g] > 1e-10):
                        pas.append(names[g])
                self.boolRules[
                    names[gp]] = ''.join(pas[:1]
                                         + [' or ' + pa for pa in pas[1:]])
                self.Adj_signed = np.sign(Coupl)
        elif self.model in ['6','7','8','9','10']:
            self.Adj_signed = np.zeros((self.dim,self.dim))
            n_sinknodes = 2
#             sinknodes = np.random.choice(np.arange(0,self.dim),
#                                              size=n_sinknodes,replace=False)
            sinknodes = np.array([0,1])
            # assume sinknodes have feeback
            self.Adj_signed[sinknodes,sinknodes] = np.ones(n_sinknodes)
#             # allow negative feedback
#             if self.model == 10:
#                 plus_minus = (np.random.randint(0,2,n_sinknodes) - 0.5)*2
#                 self.Adj_signed[sinknodes,sinknodes] = plus_minus
            leafnodes = np.array(sinknodes)
            availnodes = np.array([i for i in range(self.dim) if i not in sinknodes])
#             settings.m(0,leafnodes,availnodes)
            while len(availnodes) != 0:
                # parent
                parent_idx = np.random.choice(np.arange(0,len(leafnodes)),
                                                    size=1,replace=False)
                parent = leafnodes[parent_idx]
                # children
                children_ids = np.random.choice(np.arange(0,len(availnodes)),
                                                       size=2,replace=False)
                children = availnodes[children_ids]
                settings.m(0,parent,children)
                self.Adj_signed[children,parent] = np.ones(2)
                if self.model == 8:
                    self.Adj_signed[children[0],children[1]] = -1
                if self.model in [9,10]:
                    self.Adj_signed[children[0],children[1]] = -1
                    self.Adj_signed[children[1],children[0]] = -1
                # update leafnodes
                leafnodes = np.delete(leafnodes,parent_idx)
                leafnodes = np.append(leafnodes,children)
                # update availnodes
                availnodes = np.delete(availnodes,children_ids)
#                 settings.m(0,availnodes)
#                 settings.m(0,leafnodes)
#                 settings.m(0,self.Adj)
#                 settings.m(0,'-')
        else:
            self.Adj = np.zeros((self.dim,self.dim))
            for i in range(self.dim):
                indep = np.random.binomial(1,self.p_indep)
                if indep == 0:
                    # this number includes parents (other variables)
                    # and the variable itself, therefore its
                    # self.maxnpar+2 in the following line
                    nr = np.random.randint(1,self.maxnpar+2)
                    j_par = np.random.choice(np.arange(0,self.dim),
                                             size=nr,replace=False)
                    self.Adj[i,j_par] = 1
                else:
                    self.Adj[i,i] = 1
        #
        self.Adj = np.abs(np.array(self.Adj_signed))
        #settings.m(0,self.Adj)

    def set_coupl_old(self):
        """ Using the adjacency matrix, sample a coupling matrix.
        """
        if self.model == 'krumsiek11' or self.model == 'var':
            # we already built the coupling matrix in set_coupl20()
            return
        self.Coupl = np.zeros((self.dim,self.dim))
        for i in range(self.Adj.shape[0]):
            for j,a in enumerate(self.Adj[i]):
                # if there is a 1 in Adj, specify co and antiregulation
                # and strength of regulation
                if a != 0:
                    co_anti = np.random.randint(2)
                    # set a lower bound for the coupling parameters
                    # they ought not to be smaller than 0.1
                    # and not be larger than 0.4
                    self.Coupl[i,j] = 0.0*np.random.rand() + 0.1
                    # set sign for coupling
                    if co_anti == 1:
                        self.Coupl[i,j] *= -1
        # enforce certain requirements on models
        if self.model == 1:
            self.coupl_model1()
        elif self.model == 5:
            self.coupl_model5()
        elif self.model in [6,7]:
            self.coupl_model6()
        elif self.model in [8,9,10]:
            self.coupl_model8()
        # output
        if self.verbosity > 1:
           settings.m(0,self.Coupl)

    def coupl_model1(self):
        """ In model 1, we want enforce the following signs
            on the couplings. Model 2 has the same couplings
            but arbitrary signs.
        """
        self.Coupl[0,0] = np.abs(self.Coupl[0,0])
        self.Coupl[0,1] = -np.abs(self.Coupl[0,1])
        self.Coupl[1,1] = np.abs(self.Coupl[1,1])

    def coupl_model5(self):
        """ Toggle switch.
        """
        self.Coupl = -0.2*self.Adj
        self.Coupl[2,0] *= -1
        self.Coupl[3,0] *= -1
        self.Coupl[4,1] *= -1
        self.Coupl[5,1] *= -1

    def coupl_model6(self):
        """ Variant of toggle switch.
        """
        self.Coupl = 0.5*self.Adj_signed

    def coupl_model8(self):
        """ Variant of toggle switch.
        """
        self.Coupl = 0.5*self.Adj_signed
        # reduce the value of the coupling of the repressing features
        # otherwise completely unstable solutions are obtained
        for x in np.nditer(self.Coupl,op_flags=['readwrite']):
            if x < -1e-6:
                x[...] = -0.2

    def coupl_model_krumsiek11(self):
        """ Variant of toggle switch.
        """
        self.Coupl = self.Adj_signed

    def sim_model_back_help(self,Xt,Xt1):
        """ Yields zero when solved for X_t
            given X_{t+1}.
        """
        return - Xt1 + Xt + self.Xdiff(Xt)

    def sim_model_backwards(self,tmax,X0):
        """ Simulate the model backwards in time.
        """
        X = np.zeros((tmax,self.dim))
        X[tmax-1] = X0
        for t in range(tmax-2,-1,-1):
            sol = sp.optimize.root(self.sim_model_back_help,
                                   X[t+1],
                                   args=(X[t+1]),method='hybr')
            X[t] = sol.x
        return X

    def branch_init_model1(self,tmax=100):
        # check whether we can define trajectories
        Xfix = np.array([self.Coupl[0,1]/self.Coupl[0,0],1])
        if Xfix[0] > 0.97 or Xfix[0] < 0.03:
            settings.m(0,'... either no fixed point in [0,1]^2! \n' +
                  '    or fixed point is too close to bounds' )
            return None
        #
        XbackUp = self.sim_model_backwards(tmax=tmax/3, X0=Xfix+np.array([0.02, -0.02]))
        XbackDo = self.sim_model_backwards(tmax=tmax/3, X0=Xfix+np.array([-0.02, -0.02]))
        #
        Xup = self.sim_model(tmax=tmax, X0=XbackUp[0])
        Xdo = self.sim_model(tmax=tmax, X0=XbackDo[0])
        # compute mean
        X0mean = 0.5*(Xup[0] + Xdo[0])
        #
        if np.min(X0mean) < 0.025 or np.max(X0mean) > 0.975:
            settings.m(0, '... initial point is too close to bounds' )
            return None
        #
        if self.show and self.verbosity > 1:
            pl.figure()
            pl.plot(XbackUp[:,0],'.b',XbackUp[:,1],'.g')
            pl.plot(XbackDo[:,0],'.b',XbackDo[:,1],'.g')
            pl.plot(Xup[:,0],'b',Xup[:,1],'g')
            pl.plot(Xdo[:,0],'b',Xdo[:,1],'g')
        return X0mean

    def parents_from_boolRule(self,rule):
        """ Determine parents based on boolean updaterule.

            Returns list of parents.
        """
        rule_pa = rule.replace('(','').replace(')','').replace('or','').replace('and','').replace('not','')
        rule_pa = rule_pa.split()
        # if there are no parents, continue
        if not rule_pa:
            return []
        # check whether these are meaningful parents
        pa_old = []
        pa_delete = []
        for pa in rule_pa:
            if pa not in self.varNames.keys():
                settings.m(0,'list of available variables:')
                settings.m(0,list(self.varNames.keys()))
                message = ('processing of rule "' + rule
                             + ' yields an invalid parent: ' + pa
                             + ' | check whether the syntax is correct: \n'
                             + 'only python expressions "(",")","or","and","not" '
                             + 'are allowed, variable names and expressions have to be separated '
                             + 'by white spaces')
                raise ValueError(message)
            if pa in pa_old:
                pa_delete.append(pa)
        for pa in pa_delete:
            rule_pa.remove(pa)
        return rule_pa

    def build_boolCoeff(self):
        ''' Compute coefficients for tuple space.
        '''
        # coefficients for hill functions from boolean update rules
        self.boolCoeff = {s: [] for s in self.varNames.keys()}
        # parents
        self.pas = {s: [] for s in self.varNames.keys()}
        #
        for key in self.boolRules.keys():
            rule = self.boolRules[key]
            self.pas[key] = self.parents_from_boolRule(rule)
            pasIndices = [self.varNames[pa] for pa in self.pas[key]]
            # check whether there are coupling matrix entries for each parent
            for g in range(self.dim):
                if g in pasIndices:
                    if np.abs(self.Coupl[self.varNames[key],g]) < 1e-10:
                        raise ValueError(
                            f'specify coupling value for {key} <- {g}'
                        )
                else:
                    if np.abs(self.Coupl[self.varNames[key],g]) > 1e-10:
                        raise ValueError(
                            'there should be no coupling value for '
                            f'{key} <- {g}'
                        )
            if self.verbosity > 1:
                settings.m(0, '...'+key)
                settings.m(0, rule)
                settings.m(0, rule_pa)
            # now evaluate coefficients
            for tuple in list(itertools.product([False, True], repeat=len(self.pas[key]))):
                if self.process_rule(rule, self.pas[key], tuple):
                    self.boolCoeff[key].append(tuple)
            #
            if self.verbosity > 1:
                settings.m(0,self.boolCoeff[key])

    def process_rule(self, rule, pa, tuple):
        ''' Process a string that denotes a boolean rule.
        '''
        for i, v in enumerate(tuple):
            rule = rule.replace(pa[i], str(v))
        return eval(rule)

    def write_data(
        self,
        X,
        dir=Path('sim/test'),
        noiseObs=0.0,
        append=False,
        branching=False,
        nrRealizations=1,
        seed=0,
    ):
        header = self.header
        tmax = int(X.shape[0])
        header += 'tmax = ' + str(tmax) + '\n'
        header += 'branching = ' + str(branching) + '\n'
        header += 'nrRealizations = ' + str(nrRealizations) + '\n'
        header += 'noiseObs = ' + str(noiseObs) + '\n'
        header += 'noiseDyn = ' + str(self.noiseDyn) + '\n'
        header += 'seed = ' + str(seed) + '\n'
        # add observational noise
        X += noiseObs*np.random.randn(tmax,self.dim)
        # call helper function
        write_data(
            X,
            dir,
            append,
            header,
            varNames=self.varNames,
            Adj=self.Adj,
            Coupl=self.Coupl,
            model=self.model,
            modelType=self.modelType,
            boolRules=self.boolRules,
            invTimeStep=self.invTimeStep,
        )


def _check_branching(
    X: np.ndarray,
    Xsamples: np.ndarray,
    restart: int,
    threshold: float = 0.25
) -> Tuple[bool, List[np.ndarray]]:
    """\
    Check whether time series branches.

    Parameters
    ----------
    X
        current time series data.
    Xsamples
        list of previous branching samples.
    restart
        counts number of restart trials.
    threshold
        sets threshold for attractor identification.

    Returns
    -------
    check
        true if branching realization
    Xsamples
        updated list
    """
    check = True
    Xsamples = list(Xsamples)
    if restart == 0:
        Xsamples.append(X)
    else:
        for Xcompare in Xsamples:
            Xtmax_diff = np.absolute(X[-1, :] - Xcompare[-1, :])
            # If the second largest element is smaller than threshold
            # set check to False, i.e. at least two elements
            # need to change in order to have a branching.
            # If we observe all parameters of the system,
            # a new attractor state must involve changes in two
            # variables.
            if np.partition(Xtmax_diff, -2)[-2] < threshold:
                check = False
        if check:
            Xsamples.append(X)
    logg.debug(f'realization {restart}: {"" if check else "no"} new branch')
    return check, Xsamples


def check_nocycles(Adj: np.ndarray, verbosity: int = 2) -> bool:
    """\
    Checks that there are no cycles in graph described by adjacancy matrix.

    Parameters
    ----------
    Adj
        adjancancy matrix of dimension (dim, dim)

    Returns
    -------
    True if there is no cycle, False otherwise.
    """
    dim = Adj.shape[0]
    for g in range(dim):
        v = np.zeros(dim)
        v[g] = 1
        for i in range(dim):
            v = Adj.dot(v)
            if v[g] > 1e-10:
                if verbosity > 2:
                    settings.m(0, Adj)
                    settings.m(
                        0, 'contains a cycle of length', i+1,
                        'starting from node', g,
                        '-> reject',
                    )
                return False
    return True


def sample_coupling_matrix(
    dim: int = 3, connectivity: float = 0.5
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, int]:
    """\
    Sample coupling matrix.

    Checks that returned graphs contain no self-cycles.

    Parameters
    ----------
    dim
        dimension of coupling matrix.
    connectivity
        fraction of connectivity, fully connected means 1.,
        not-connected means 0, in the case of fully connected, one has
        dim*(dim-1)/2 edges in the graph.

    Returns
    -------
    coupl
        coupling matrix
    adj
        adjancancy matrix
    adj_signed
        signed adjacancy matrix
    n_edges
        Number of edges
    """
    max_trial = 10
    check = False
    for trial in range(max_trial):
        # random topology for a given connectivity / edge density
        Coupl = np.zeros((dim, dim))
        n_edges = 0
        for gp in range(dim):
            for g in range(dim):
                if gp == g:
                    continue
                # need to have the factor 0.5, otherwise
                # connectivity=1 would lead to dim*(dim-1) edges
                if np.random.rand() < 0.5*connectivity:
                    Coupl[gp, g] = 0.7
                    n_edges += 1
        # obtain adjacancy matrix
        Adj_signed = np.zeros((dim, dim), dtype='int_')
        Adj_signed = np.sign(Coupl)
        Adj = np.abs(Adj_signed)
        # check for cycles and whether there is at least one edge
        if check_nocycles(Adj) and n_edges > 0:
            check = True
            break
    if not check:
        raise ValueError(
            'did not find graph without cycles after'
            f'{max_trial} trials'
        )
    return Coupl, Adj, Adj_signed, n_edges


class StaticCauseEffect:
    """
    Simulates static data to investigate structure learning.
    """

    availModels = dict(
        line='y = αx \n',
        noise='y = noise \n',
        absline='y = |x| \n',
        parabola='y = αx² \n',
        sawtooth='y = x - |x| \n',
        tanh='y = tanh(x) \n',
        combi='combinatorial regulation \n',
    )

    def __init__(self):
        # define a set of available functions
        self.funcs = dict(
            line=lambda x: x,
            noise=lambda x: 0,
            absline=np.abs,
            parabola=lambda x: x ** 2,
            sawtooth=lambda x: 0.5 * x - np.floor(0.5 * x),
            tanh=lambda x: np.tanh(2 * x),
        )

    def sim_givenAdj(self, Adj: np.ndarray, model='line'):
        """\
        Simulate data given only an adjacancy matrix and a model.

        The model is a bivariate funtional dependence. The adjacancy matrix
        needs to be acyclic.

        Parameters
        ----------
        Adj
            adjacancy matrix of shape (dim,dim).

        Returns
        -------
        Data array of shape (n_samples,dim).
        """
        # nice examples
        examples = [dict(
            func='sawtooth',
            gdist='uniform',
            sigma_glob=1.8,
            sigma_noise=0.1,
        )]

        # nr of samples
        n_samples = 100

        # noise
        sigma_glob = 1.8
        sigma_noise = 0.4

        # coupling function / model
        func = self.funcs[model]

        # glob distribution
        sourcedist = 'uniform'

        # loop over source nodes
        dim = Adj.shape[0]
        X = np.zeros((n_samples, dim))
        # source nodes have no parents themselves
        nrpar = 0
        children = list(range(dim))
        parents = []
        for gp in range(dim):
            if Adj[gp, :].sum() == nrpar:
                if sourcedist == 'gaussian':
                    X[:, gp] = np.random.normal(0, sigma_glob, n_samples)
                if sourcedist == 'uniform':
                    X[:, gp] = np.random.uniform(-sigma_glob, sigma_glob, n_samples)
                parents.append(gp)
                children.remove(gp)

        # all of the following guarantees for 3 dim, that we generate the data
        # in the correct sequence
        # then compute all nodes that have 1 parent, then those with 2 parents
        children_sorted = []
        nrchildren_par = np.zeros(dim)
        nrchildren_par[0] = len(parents)
        for nrpar in range(1, dim):
            # loop over child nodes
            for gp in children:
                if Adj[gp, :].sum() == nrpar:
                    children_sorted.append(gp)
                    nrchildren_par[nrpar] += 1
        # if there is more than a child with a single parent
        # order these children (there are two in three dim)
        # by distance to the source/parent
        if nrchildren_par[1] > 1:
            if Adj[children_sorted[0], parents[0]] == 0:
                help = children_sorted[0]
                children_sorted[0] = children_sorted[1]
                children_sorted[1] = help

        for gp in children_sorted:
            for g in range(dim):
                if Adj[gp, g] > 0:
                    X[:, gp] += 1./Adj[gp, :].sum()*func(X[:, g])
            X[:, gp] += np.random.normal(0, sigma_noise, n_samples)

#         fig = pl.figure()
#         fig.add_subplot(311)
#         pl.plot(X[:,0],X[:,1],'.',mec='white')
#         fig.add_subplot(312)
#         pl.plot(X[:,1],X[:,2],'.',mec='white')
#         fig.add_subplot(313)
#         pl.plot(X[:,2],X[:,0],'.',mec='white')
#         pl.show()

        return X

    def sim_combi(self):
        """ Simulate data to model combi regulation.
        """
        n_samples = 500
        sigma_glob = 1.8

        X = np.zeros((n_samples, 3))

        X[:, 0] = np.random.uniform(-sigma_glob, sigma_glob, n_samples)
        X[:, 1] = np.random.uniform(-sigma_glob, sigma_glob, n_samples)

        func = self.funcs['tanh']

        # XOR type
#         X[:,2] = (func(X[:,0])*sp.stats.norm.pdf(X[:,1],0,0.2)
#                   + func(X[:,1])*sp.stats.norm.pdf(X[:,0],0,0.2))
        # AND type / diagonal
#         X[:,2] = (func(X[:,0]+X[:,1])*sp.stats.norm.pdf(X[:,1]-X[:,0],0,0.2))
        # AND type / horizontal
        X[:, 2] = (func(X[:, 0])*sp.stats.norm.cdf(X[:, 1], 1, 0.2))

        pl.scatter(X[:, 0], X[:, 1], c=X[:, 2], edgecolor='face')
        pl.show()

        pl.plot(X[:, 1], X[:, 2], '.')
        pl.show()

        return X


def sample_static_data(model, dir, verbosity=0):
    # fraction of connectivity as compared to fully connected
    # in one direction, which amounts to dim*(dim-1)/2 edges
    connectivity = 0.8
    dim = 3
    n_Coupls = 50
    model = model.replace('static-', '')
    np.random.seed(0)

    if model != 'combi':
        n_edges = np.zeros(n_Coupls)
        for icoupl in range(n_Coupls):
            Coupl, Adj, Adj_signed, n_e = sample_coupling_matrix(dim,connectivity)
            if verbosity > 1:
                settings.m(0,icoupl)
                settings.m(0,Adj)
            n_edges[icoupl] = n_e
            # sample data
            X = StaticCauseEffect().sim_givenAdj(Adj,model)
            write_data(X, dir, Adj=Adj)
        settings.m(0, 'mean edge number:', n_edges.mean())

    else:
        X = StaticCauseEffect().sim_combi()
        Adj = np.zeros((3, 3))
        Adj[2, 0] = Adj[2, 1] = 0
        write_data(X, dir, Adj=Adj)


if __name__ == '__main__':
    import argparse
#     epilog = ('    1: 2dim, causal direction X_1 -> X_0, constraint signs\n'
#               + '    2: 2dim, causal direction X_1 -> X_0, arbitrary signs\n'
#               + '    3: 2dim, causal direction X_1 <-> X_0, arbitrary signs\n'
#               + '    4: 2dim, mix of model 2 and 3\n'
#               + '    5: 6dim double toggle switch\n'
#               + '    6: two independent evolutions without repression, sync.\n'
#               + '    7: two independent evolutions without repression, random init\n'
#               + '    8: two independent evolutions directed repression, random init\n'
#               + '    9: two independent evolutions mutual repression, random init\n'
#               + '   10: two indep. evol., diff. self-loops possible, mut. repr., rand init\n')
    epilog = ''
    for k,v in StaticCauseEffect.availModels.items():
        epilog += '    static-' + k + ': ' + v
    for k,v in GRNsim.availModels.items():
        epilog += '    ' + k + ': ' + v
    # command line options
    p = argparse.ArgumentParser(
        description=('Simulate stochastic discrete-time dynamical systems,\n'
                     'in particular feature regulatory networks.'),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=('  MODEL: specify one of the following models, or one of \n'
                '    the filenames (without ".txt") in the directory "models" \n'
                + epilog
                ))
    aa = p.add_argument
    dir_arg = aa('--dir',required=True,
        type=str,default='',
        help=('specify directory to store data, '
              + ' must start with "sim/MODEL_...", see possible values for MODEL below '))
    aa('--show',
        action='store_true',
        help='show plots')
    aa('--verbosity',
        type=int,default=0,
        help='specify integer > 0 to get more output [default 0]')
    args = p.parse_args()

    # run checks on output directory
    dir = Path(args.dir)
    if not dir.resolve().parent.name == 'sim':
        raise argparse.ArgumentError(
            dir_arg,
            "The parent directory of the --dir argument needs to be named 'sim'"
        )
    else:
        model = dir.name.split('_')[0]
        settings.m(0, f'...model is: {model!r}')
    if dir.is_dir() and 'test' not in str(dir):
        message = (
            f'directory {dir} already exists, '
            'remove it and continue? [y/n, press enter]'
        )
        if str(input(message)) != 'y':
            settings.m(0, '    ...quit program execution')
            sys.exit()
        else:
            settings.m(0, '   ...removing directory and continuing...')
            shutil.rmtree(dir)

    settings.m(0, model)
    settings.m(0, dir)

    # sample data
    if 'static' in model:
        sample_static_data(model=model, dir=dir, verbosity=args.verbosity)
    else:
        sample_dynamic_data(model=model, dir=dir)



"""Reading and Writing
"""
from pathlib import Path, PurePath
from typing import Union, Dict, Optional, Tuple, BinaryIO

import json
import numpy as np
import pandas as pd
from matplotlib.image import imread
import tables
import anndata
from anndata import (
    AnnData,
    read_csv,
    read_text,
    read_excel,
    read_mtx,
    read_loom,
    read_hdf,
)
from anndata import read as read_h5ad

from ._settings import settings
from ._compat import Literal
from ._utils import Empty, _empty
from . import logging as logg

# .gz and .bz2 suffixes are also allowed for text formats
text_exts = {
    'csv',
    'tsv',
    'tab',
    'data',
    'txt',  # these four are all equivalent
}
avail_exts = {
    'anndata',
    'xlsx',
    'h5',
    'h5ad',
    'mtx',
    'mtx.gz',
    'soft.gz',
    'loom',
} | text_exts
"""Available file formats for reading data. """


# --------------------------------------------------------------------------------
# Reading and Writing data files and AnnData objects
# --------------------------------------------------------------------------------


def read(
    filename: Union[Path, str],
    backed: Optional[Literal['r', 'r+']] = None,
    sheet: Optional[str] = None,
    ext: Optional[str] = None,
    delimiter: Optional[str] = None,
    first_column_names: bool = False,
    backup_url: Optional[str] = None,
    cache: bool = False,
    cache_compression: Union[Literal['gzip', 'lzf'], None, Empty] = _empty,
    **kwargs) -> AnnData:
    """\
    Read file and return :class:`~anndata.AnnData` object.
    To speed up reading, consider passing ``cache=True``, which creates an hdf5
    cache file.
    Parameters
    ----------
    filename
        If the filename has no file extension, it is interpreted as a key for
        generating a filename via ``sc.settings.writedir / (filename +
        sc.settings.file_format_data)``.  This is the same behavior as in
        ``sc.read(filename, ...)``.
    backed
        If ``'r'``, load :class:`~anndata.AnnData` in ``backed`` mode instead
        of fully loading it into memory (`memory` mode). If you want to modify
        backed attributes of the AnnData object, you need to choose ``'r+'``.
    sheet
        Name of sheet/table in hdf5 or Excel file.
    ext
        Extension that indicates the file type. If ``None``, uses extension of
        filename.
    delimiter
        Delimiter that separates data within text file. If ``None``, will split at
        arbitrary number of white spaces, which is different from enforcing
        splitting at any single white space ``' '``.
    first_column_names
        Assume the first column stores row names. This is only necessary if
        these are not strings: strings in the first column are automatically
        assumed to be row names.
    backup_url
        Retrieve the file from an URL if not present on disk.
    cache
        If `False`, read from source, if `True`, read from fast 'h5ad' cache.
    cache_compression
        See the h5py :ref:`dataset_compression`.
        (Default: `settings.cache_compression`)
    kwargs
        Parameters passed to :func:`~anndata.read_loom`.
    Returns
    -------
    An :class:`~anndata.AnnData` object
    """
    filename = Path(filename)  # allow passing strings
    if is_valid_filename(filename):
        return _read(
            filename,
            backed=backed,
            sheet=sheet,
            ext=ext,
            delimiter=delimiter,
            first_column_names=first_column_names,
            backup_url=backup_url,
            cache=cache,
            cache_compression=cache_compression,
            **kwargs,
        )
    # generate filename and read to dict
    filekey = str(filename)
    filename = settings.writedir / (filekey + '.' + settings.file_format_data)
    if not filename.exists():
        raise ValueError(
            'Reading with filekey {} failed, \
            the inferred filename {} does not exist. \
            If you intended to provide a filename, either use a filename \
            ending on one of the available extensions {} \
            or pass the parameter `ext`.'.format(filekey, filename, avail_exts)
        )
    return read_h5ad(filename, backed=backed)


# def read_10x_h5(
#     filename: Union[str, Path],
#     genome: Optional[str] = None,
#     gex_only: bool = True,
#     backup_url: Optional[str] = None) -> AnnData:
#     """\
#     Read 10x-Genomics-formatted hdf5 file.
#     Parameters
#     ----------
#     filename
#         Path to a 10x hdf5 file.
#     genome
#         Filter expression to genes within this genome. For legacy 10x h5
#         files, this must be provided if the data contains more than one genome.
#     gex_only
#         Only keep 'Gene Expression' data and ignore other feature types,
#         e.g. 'Antibody Capture', 'CRISPR Guide Capture', or 'Custom'
#     backup_url
#         Retrieve the file from an URL if not present on disk.
#     Returns
#     -------
#     Annotated data matrix, where observations/cells are named by their
#     barcode and variables/genes by gene name. Stores the following information:
#     :attr:`~anndata.AnnData.X`
#         The data matrix is stored
#     :attr:`~anndata.AnnData.obs_names`
#         Cell names
#     :attr:`~anndata.AnnData.var_names`
#         Gene names
#     :attr:`~anndata.AnnData.var`\\ `['gene_ids']`
#         Gene IDs
#     :attr:`~anndata.AnnData.var`\\ `['feature_types']`
#         Feature types
#     """
#     start = logg.info(f'reading {filename}')
#     is_present = _check_datafile_present_and_download(filename, backup_url=backup_url,)
#     if not is_present:
#         logg.debug(f'... did not find original file {filename}')
#     with tables.open_file(str(filename), 'r') as f:
#         v3 = '/matrix' in f
#     if v3:
#         adata = _read_v3_10x_h5(filename, start=start)
#         if genome:
#             if genome not in adata.var['genome'].values:
#                 raise ValueError(
#                     f"Could not find data corresponding to genome '{genome}' in '{filename}'. "
#                     f'Available genomes are: {list(adata.var["genome"].unique())}.'
#                 )
#             adata = adata[:, list(map(lambda x: x == str(genome), adata.var['genome']))]
#         if gex_only:
#             adata = adata[
#                 :,
#                 list(map(lambda x: x == 'Gene Expression', adata.var['feature_types'])),
#             ]
#         if adata.is_view:
#             adata = adata.copy()
#     else:
#         adata = _read_legacy_10x_h5(filename, genome=genome, start=start)
#     return adata


# def _read_legacy_10x_h5(filename, *, genome=None, start=None):
#     """
#     Read hdf5 file from Cell Ranger v2 or earlier versions.
#     """
#     with tables.open_file(str(filename), 'r') as f:
#         try:
#             children = [x._v_name for x in f.list_nodes(f.root)]
#             if not genome:
#                 if len(children) > 1:
#                     raise ValueError(
#                         f"'{filename}' contains more than one genome. For legacy 10x h5 "
#                         "files you must specify the genome if more than one is present. "
#                         f"Available genomes are: {children}"
#                     )
#                 genome = children[0]
#             elif genome not in children:
#                 raise ValueError(
#                     f"Could not find genome '{genome}' in '{filename}'. "
#                     f'Available genomes are: {children}'
#                 )
#             dsets = {}
#             for node in f.walk_nodes('/' + genome, 'Array'):
#                 dsets[node.name] = node.read()
#             # AnnData works with csr matrices
#             # 10x stores the transposed data, so we do the transposition right away
#             from scipy.sparse import csr_matrix

#             M, N = dsets['shape']
#             data = dsets['data']
#             if dsets['data'].dtype == np.dtype('int32'):
#                 data = dsets['data'].view('float32')
#                 data[:] = dsets['data']
#             matrix = csr_matrix(
#                 (data, dsets['indices'], dsets['indptr']), shape=(N, M),
#             )
#             # the csc matrix is automatically the transposed csr matrix
#             # as scanpy expects it, so, no need for a further transpostion
#             adata = AnnData(
#                 matrix,
#                 dict(obs_names=dsets['barcodes'].astype(str)),
#                 dict(
#                     var_names=dsets['gene_names'].astype(str),
#                     gene_ids=dsets['genes'].astype(str),
#                 ),
#             )
#             logg.info('', time=start)
#             return adata
#         except KeyError:
#             raise Exception('File is missing one or more required datasets.')


# def _read_v3_10x_h5(filename, *, start=None):
#     """
#     Read hdf5 file from Cell Ranger v3 or later versions.
#     """
#     with tables.open_file(str(filename), 'r') as f:
#         try:
#             dsets = {}
#             for node in f.walk_nodes('/matrix', 'Array'):
#                 dsets[node.name] = node.read()
#             from scipy.sparse import csr_matrix

#             M, N = dsets['shape']
#             data = dsets['data']
#             if dsets['data'].dtype == np.dtype('int32'):
#                 data = dsets['data'].view('float32')
#                 data[:] = dsets['data']
#             matrix = csr_matrix(
#                 (data, dsets['indices'], dsets['indptr']), shape=(N, M),
#             )
#             adata = AnnData(
#                 matrix,
#                 dict(obs_names=dsets['barcodes'].astype(str)),
#                 dict(
#                     var_names=dsets['name'].astype(str),
#                     gene_ids=dsets['id'].astype(str),
#                     feature_types=dsets['feature_type'].astype(str),
#                     genome=dsets['genome'].astype(str),
#                 ),
#             )
#             logg.info('', time=start)
#             return adata
#         except KeyError:
#             raise Exception('File is missing one or more required datasets.')




# def read_10x_mtx(
#     path: Union[Path, str],
#     var_names: Literal['gene_symbols', 'gene_ids'] = 'gene_symbols',
#     make_unique: bool = True,
#     cache: bool = False,
#     cache_compression: Union[Literal['gzip', 'lzf'], None, Empty] = _empty,
#     gex_only: bool = True,
#     *,
#     prefix: str = None) -> AnnData:
#     """\
#     Read 10x-Genomics-formatted mtx directory.
#     Parameters
#     ----------
#     path
#         Path to directory for `.mtx` and `.tsv` files,
#         e.g. './filtered_gene_bc_matrices/hg19/'.
#     var_names
#         The variables index.
#     make_unique
#         Whether to make the variables index unique by appending '-1',
#         '-2' etc. or not.
#     cache
#         If `False`, read from source, if `True`, read from fast 'h5ad' cache.
#     cache_compression
#         See the h5py :ref:`dataset_compression`.
#         (Default: `settings.cache_compression`)
#     gex_only
#         Only keep 'Gene Expression' data and ignore other feature types,
#         e.g. 'Antibody Capture', 'CRISPR Guide Capture', or 'Custom'
#     prefix
#         Any prefix before `matrix.mtx`, `genes.tsv` and `barcodes.tsv`. For instance,
#         if the files are named `patientA_matrix.mtx`, `patientA_genes.tsv` and
#         `patientA_barcodes.tsv` the prefix is `patientA_`.
#         (Default: no prefix)
#     Returns
#     -------
#     An :class:`~anndata.AnnData` object
#     """
#     path = Path(path)
#     prefix = "" if prefix is None else prefix
#     genefile_exists = (path / f'{prefix}genes.tsv').is_file()
#     read = _read_legacy_10x_mtx if genefile_exists else _read_v3_10x_mtx
#     adata = read(
#         str(path),
#         var_names=var_names,
#         make_unique=make_unique,
#         cache=cache,
#         cache_compression=cache_compression,
#         prefix=prefix)
#     if genefile_exists or not gex_only:
#         return adata
#     else:
#         gex_rows = list(
#             map(lambda x: x == 'Gene Expression', adata.var['feature_types'])
#         )
#         return adata[:, gex_rows].copy()


# def _read_legacy_10x_mtx(
#     path,
#     var_names='gene_symbols',
#     make_unique=True,
#     cache=False,
#     cache_compression=_empty,
#     *,
#     prefix=""):
#     """
#     Read mex from output from Cell Ranger v2 or earlier versions
#     """
#     path = Path(path)
#     adata = read(
#         path / f'{prefix}matrix.mtx', cache=cache, cache_compression=cache_compression,
#     ).T  # transpose the data
#     genes = pd.read_csv(path / f'{prefix}genes.tsv', header=None, sep='\t')
#     if var_names == 'gene_symbols':
#         var_names = genes[1].values
#         if make_unique:
#             var_names = anndata.utils.make_index_unique(pd.Index(var_names))
#         adata.var_names = var_names
#         adata.var['gene_ids'] = genes[0].values
#     elif var_names == 'gene_ids':
#         adata.var_names = genes[0].values
#         adata.var['gene_symbols'] = genes[1].values
#     else:
#         raise ValueError("`var_names` needs to be 'gene_symbols' or 'gene_ids'")
#     adata.obs_names = pd.read_csv(path / f'{prefix}barcodes.tsv', header=None)[0].values
#     return adata


# def _read_v3_10x_mtx(
#     path,
#     var_names='gene_symbols',
#     make_unique=True,
#     cache=False,
#     cache_compression=_empty,
#     *,
#     prefix=""):
#     """
#     Read mtx from output from Cell Ranger v3 or later versions
#     """
#     path = Path(path)
#     adata = read(
#         path / f'{prefix}matrix.mtx.gz',
#         cache=cache,
#         cache_compression=cache_compression,
#     ).T  # transpose the data
#     genes = pd.read_csv(path / f'{prefix}features.tsv.gz', header=None, sep='\t')
#     if var_names == 'gene_symbols':
#         var_names = genes[1].values
#         if make_unique:
#             var_names = anndata.utils.make_index_unique(pd.Index(var_names))
#         adata.var_names = var_names
#         adata.var['gene_ids'] = genes[0].values
#     elif var_names == 'gene_ids':
#         adata.var_names = genes[0].values
#         adata.var['gene_symbols'] = genes[1].values
#     else:
#         raise ValueError("`var_names` needs to be 'gene_symbols' or 'gene_ids'")
#     adata.var['feature_types'] = genes[2].values
#     adata.obs_names = pd.read_csv(path / f'{prefix}barcodes.tsv.gz', header=None)[
#         0
#     ].values
#     return adata


def write(
    filename: Union[str, Path],
    adata: AnnData,
    ext: Optional[Literal['h5', 'csv', 'txt', 'npz']] = None,
    compression: Optional[Literal['gzip', 'lzf']] = 'gzip',
    compression_opts: Optional[int] = None,
):
    """\
    Write :class:`~anndata.AnnData` objects to file.
    Parameters
    ----------
    filename
        If the filename has no file extension, it is interpreted as a key for
        generating a filename via `sc.settings.writedir / (filename +
        sc.settings.file_format_data)`. This is the same behavior as in
        :func:`~scanpy.read`.
    adata
        Annotated data matrix.
    ext
        File extension from wich to infer file format. If `None`, defaults to
        `sc.settings.file_format_data`.
    compression
        See http://docs.h5py.org/en/latest/high/dataset.html.
    compression_opts
        See http://docs.h5py.org/en/latest/high/dataset.html.
    """
    filename = Path(filename)  # allow passing strings
    if is_valid_filename(filename):
        filename = filename
        ext_ = is_valid_filename(filename, return_ext=True)
        if ext is None:
            ext = ext_
        elif ext != ext_:
            raise ValueError(
                'It suffices to provide the file type by '
                'providing a proper extension to the filename.'
                'One of "txt", "csv", "h5" or "npz".'
            )
    else:
        key = filename
        ext = settings.file_format_data if ext is None else ext
        filename = _get_filename_from_key(key, ext)
    if ext == 'csv':
        adata.write_csvs(filename)
    else:
        adata.write(
            filename, compression=compression, compression_opts=compression_opts
        )


# -------------------------------------------------------------------------------
# Reading and writing parameter files
# -------------------------------------------------------------------------------


def read_params(
    filename: Union[Path, str], asheader: bool = False) -> Dict[str, Union[int, float, bool, str, None]]:
    """\
    Read parameter dictionary from text file.
    Assumes that parameters are specified in the format::
        par1 = value1
        par2 = value2
    Comments that start with '#' are allowed.
    Parameters
    ----------
    filename
        Filename of data file.
    asheader
        Read the dictionary from the header (comment section) of a file.
    Returns
    -------
    Dictionary that stores parameters.
    """
    filename = str(filename)  # allow passing pathlib.Path objects
    from collections import OrderedDict

    params = OrderedDict([])
    for line in open(filename):
        if '=' in line:
            if not asheader or line.startswith('#'):
                line = line[1:] if line.startswith('#') else line
                key, val = line.split('=')
                key = key.strip()
                val = val.strip()
                params[key] = convert_string(val)
    return params


def write_params(path: Union[Path, str], *args, **maps):
    """\
    Write parameters to file, so that it's readable by read_params.
    Uses INI file format.
    """
    path = Path(path)
    if not path.parent.is_dir():
        path.parent.mkdir(parents=True)
    if len(args) == 1:
        maps[None] = args[0]
    with path.open('w') as f:
        for header, map in maps.items():
            if header is not None:
                f.write('[{}]\n'.format(header))
            for key, val in map.items():
                f.write('{} = {}\n'.format(key, val))


# -------------------------------------------------------------------------------
# Reading and Writing data files
# -------------------------------------------------------------------------------


def _read(
    filename: Path,
    backed=None,
    sheet=None,
    ext=None,
    delimiter=None,
    first_column_names=None,
    backup_url=None,
    cache=False,
    cache_compression=None,
    suppress_cache_warning=False,
    **kwargs):
    if ext is not None and ext not in avail_exts:
        raise ValueError(
            'Please provide one of the available extensions.\n{}'.format(avail_exts)
        )
    else:
        ext = is_valid_filename(filename, return_ext=True)
    is_present = _check_datafile_present_and_download(filename, backup_url=backup_url,)
    if not is_present:
        logg.debug('... did not find original file {}'.format(filename))
    # read hdf5 files
    if ext in {'h5', 'h5ad'}:
        if sheet is None:
            return read_h5ad(filename, backed=backed)
        else:
            logg.debug('reading sheet {} from file {}'.format(sheet, filename))
            return read_hdf(filename, sheet)
    # read other file types
    path_cache = settings.cachedir / _slugify(filename).replace(
        '.' + ext, '.h5ad'
    )  # type: Path
    if path_cache.suffix in {'.gz', '.bz2'}:
        path_cache = path_cache.with_suffix('')
    if cache and path_cache.is_file():
        logg.info('... reading from cache file {}'.format(path_cache))
        return read_h5ad(path_cache)

    if not is_present:
        raise FileNotFoundError('Did not find file {}.'.format(filename))
    logg.debug('reading {}'.format(filename))
    if not cache and not suppress_cache_warning:
        logg.hint(
            'This might be very slow. Consider passing `cache=True`, '
            'which enables much faster reading from a cache file.'
        )
    # do the actual reading
    if ext == 'xlsx' or ext == 'xls':
        if sheet is None:
            raise ValueError("Provide `sheet` parameter when reading '.xlsx' files.")
        else:
            adata = read_excel(filename, sheet)
    elif ext in {'mtx', 'mtx.gz'}:
        adata = read_mtx(filename)
    elif ext == 'csv':
        adata = read_csv(filename, first_column_names=first_column_names)
    elif ext in {'txt', 'tab', 'data', 'tsv'}:
        if ext == 'data':
            logg.hint(
                "... assuming '.data' means tab or white-space " 'separated text file',
            )
            logg.hint('change this by passing `ext` to sc.read')
        adata = read_text(filename, delimiter, first_column_names)
    elif ext == 'soft.gz':
        adata = _read_softgz(filename)
    elif ext == 'loom':
        adata = read_loom(filename=filename, **kwargs)
    else:
        raise ValueError('Unknown extension {}.'.format(ext))
    if cache:
        logg.info(
            '... writing an {} cache file to speedup \
                reading next time'.format(settings.file_format_data)
        )
        if cache_compression is _empty:
            cache_compression = settings.cache_compression
        if not path_cache.parent.is_dir():
            path_cache.parent.mkdir(parents=True)
        # write for faster reading when calling the next time
        adata.write(path_cache, compression=cache_compression)
    return adata


def _slugify(path: Union[str, PurePath]) -> str:
    """Make a path into a filename."""
    if not isinstance(path, PurePath):
        path = PurePath(path)
    parts = list(path.parts)
    if parts[0] == '/':
        parts.pop(0)
    elif len(parts[0]) == 3 and parts[0][1:] == ':\\':
        parts[0] = parts[0][0]  # C:\ → C
    filename = '-'.join(parts)
    assert '/' not in filename, filename
    assert not filename[1:].startswith(':'), filename
    return filename


def _read_softgz(filename: Union[str, bytes, Path, BinaryIO]) -> AnnData:
    """\
    Read a SOFT format data file.
    The SOFT format is documented here
    http://www.ncbi.nlm.nih.gov/geo/info/soft2.html.
    Notes
    -----
    The function is based on a script by Kerby Shedden.
    http://dept.stat.lsa.umich.edu/~kshedden/Python-Workshop/gene_expression_comparison.html
    """
    import gzip

    with gzip.open(filename, mode='rt') as file:
        # The header part of the file contains information about the
        # samples. Read that information first.
        samples_info = {}
        for line in file:
            if line.startswith("!dataset_table_begin"):
                break
            elif line.startswith("!subset_description"):
                subset_description = line.split("=")[1].strip()
            elif line.startswith("!subset_sample_id"):
                subset_ids = line.split("=")[1].split(",")
                subset_ids = [x.strip() for x in subset_ids]
                for k in subset_ids:
                    samples_info[k] = subset_description
        # Next line is the column headers (sample id's)
        sample_names = file.readline().strip().split("\t")
        # The column indices that contain gene expression data
        I = [i for i, x in enumerate(sample_names) if x.startswith("GSM")]
        # Restrict the column headers to those that we keep
        sample_names = [sample_names[i] for i in I]
        # Get a list of sample labels
        groups = [samples_info[k] for k in sample_names]
        # Read the gene expression data as a list of lists, also get the gene
        # identifiers
        gene_names, X = [], []
        for line in file:
            # This is what signals the end of the gene expression data
            # section in the file
            if line.startswith("!dataset_table_end"):
                break
            V = line.split("\t")
            # Extract the values that correspond to gene expression measures
            # and convert the strings to numbers
            x = [float(V[i]) for i in I]
            X.append(x)
            gene_names.append(V[1])
    # Convert the Python list of lists to a Numpy array and transpose to match
    # the Scanpy convention of storing samples in rows and variables in colums.
    X = np.array(X).T
    obs = pd.DataFrame({"groups": groups}, index=sample_names)
    var = pd.DataFrame(index=gene_names)
    return AnnData(X=X, obs=obs, var=var)


# -------------------------------------------------------------------------------
# Type conversion
# -------------------------------------------------------------------------------


def is_float(string: str) -> float:
    """Check whether string is float.
    See also
    --------
    http://stackoverflow.com/questions/736043/checking-if-a-string-can-be-converted-to-float-in-python
    """
    try:
        float(string)
        return True
    except ValueError:
        return False


def is_int(string: str) -> bool:
    """Check whether string is integer.
    """
    try:
        int(string)
        return True
    except ValueError:
        return False


def convert_bool(string: str) -> Tuple[bool, bool]:
    """Check whether string is boolean.
    """
    if string == 'True':
        return True, True
    elif string == 'False':
        return True, False
    else:
        return False, False


def convert_string(string: str) -> Union[int, float, bool, str, None]:
    """Convert string to int, float or bool.
    """
    if is_int(string):
        return int(string)
    elif is_float(string):
        return float(string)
    elif convert_bool(string)[0]:
        return convert_bool(string)[1]
    elif string == 'None':
        return None
    else:
        return string


# -------------------------------------------------------------------------------
# Helper functions for reading and writing
# -------------------------------------------------------------------------------


def get_used_files():
    """Get files used by processes with name scanpy."""
    import psutil

    loop_over_scanpy_processes = (
        proc for proc in psutil.process_iter() if proc.name() == 'scanpy'
    )
    filenames = []
    for proc in loop_over_scanpy_processes:
        try:
            flist = proc.open_files()
            for nt in flist:
                filenames.append(nt.path)
        # This catches a race condition where a process ends
        # before we can examine its files
        except psutil.NoSuchProcess as err:
            pass
    return set(filenames)


def _get_filename_from_key(key, ext=None) -> Path:
    ext = settings.file_format_data if ext is None else ext
    return settings.writedir / '{key}.{ext}'.format(key, ext)


def _download(url: str, path: Path):
    try:
        import ipywidgets
        from tqdm.auto import tqdm
    except ModuleNotFoundError:
        from tqdm import tqdm
    from urllib.request import urlopen, Request

    blocksize = 1024 * 8
    blocknum = 0

    try:
        # This is a modified urllib.request.urlretrieve, but that won't let us
        # set headers.
        # The '\'s are ugly, but parenthesis are not allowed here
        # fmt: off
        with \
            tqdm(unit='B', unit_scale=True, miniters=1, desc=path.name) as t, \
            path.open("wb") as f, \
            urlopen(Request(url, headers={"User-agent": "scanpy-user"})) as resp:

            t.total = int(resp.info().get("content-length", 0))

            block = resp.read(blocksize)
            while block:
                f.write(block)
                blocknum += 1
                t.update(blocknum * blocksize - t.n)
                block = resp.read(blocksize)
        # fmt: on
    except Exception:
        # Make sure file doesn’t exist half-downloaded
        if path.is_file():
            path.unlink()
        raise


def _check_datafile_present_and_download(path, backup_url=None):
    """Check whether the file is present, otherwise download.
    """
    path = Path(path)
    if path.is_file():
        return True
    if backup_url is None:
        return False
    logg.info(
        'try downloading from url\n{}\n \
        ... this may take a while but only happens once'.format(backup_url)
    )
    if not path.parent.is_dir():
        logg.info('creating directory {}/ for saving data'.format(path.parent))
        path.parent.mkdir(parents=True)

    _download(backup_url, path)
    return True


def is_valid_filename(filename: Path, return_ext=False):
    """Check whether the argument is a filename."""
    ext = filename.suffixes

    if len(ext) > 2:
        logg.warning(
            'Your filename has more than two extensions: {ext}.\n \
            Only considering the two last: {}.'.format(ext, ext[-2:])
        )
        ext = ext[-2:]

    # cases for gzipped/bzipped text files
    if len(ext) == 2 and ext[0][1:] in text_exts and ext[1][1:] in ('gz', 'bz2'):
        return ext[0][1:] if return_ext else True
    elif ext and ext[-1][1:] in avail_exts:
        return ext[-1][1:] if return_ext else True
    elif ''.join(ext) == '.soft.gz':
        return 'soft.gz' if return_ext else True
    elif ''.join(ext) == '.mtx.gz':
        return 'mtx.gz' if return_ext else True
    elif not return_ext:
        return False
    raise ValueError(
        '''\
{} does not end on a valid extension.
Please, provide one of the available extensions.
{}
Text files with .gz and .bz2 extensions are also supported.\
'''.format(filename, avail_exts)
    )
